"""Home to various lifetime models, for use in dynamic stock modelling."""

from abc import abstractmethod
import numpy as np
import scipy.stats
from pydantic import BaseModel as PydanticBaseModel, model_validator
from typing import Any

# from scipy.special import gammaln, logsumexp
# from scipy.optimize import root_scalar

from .dimensions import DimensionSet
from .flodym_arrays import FlodymArray


class LifetimeModel(PydanticBaseModel):
    """Contains shared functionality across the various lifetime models."""

    dims: DimensionSet
    time_letter: str = "t"
    _sf: np.ndarray = None

    @property
    def t(self):
        return np.array(self.dims[self.time_letter].items)

    @property
    def shape(self):
        return self.dims.shape()

    @property
    def _n_t(self):
        return len(self.t)

    @property
    def _shape_cohort(self):
        return (self._n_t,) + self.shape

    @property
    def _shape_no_t(self):
        return tuple(list(self.shape)[1:])

    @property
    def sf(self):
        if self._sf is None:
            self._sf = np.zeros(self._shape_cohort)
            self.compute_survival_factor()
        return self._sf

    @property
    def t_diag_indices(self):
        return np.diag_indices(self._n_t) + (slice(None),) * len(self._shape_no_t)

    def _tile(self, a: np.ndarray) -> np.ndarray:
        index = (slice(None),) * a.ndim + (np.newaxis,) * len(self._shape_no_t)
        out = a[index]
        return np.tile(out, self._shape_no_t)

    def _remaining_ages(self, m):
        return self._tile(self.t[m:] - self.t[m])

    def compute_survival_factor(self):
        """Survival table self.sf(m,n) denotes the share of an inflow in year n (age-cohort) still
        present at the end of year m (after m-n years).
        The computation is self.sf(m,n) = ProbDist.sf(m-n), where ProbDist is the appropriate
        scipy function for the lifetime model chosen.
        For lifetimes 0 the sf is also 0, meaning that the age-cohort leaves during the same year
        of the inflow.
        The method compute outflow_sf returns an array year-by-cohort of the surviving fraction of
        a flow added to stock in year m (aka cohort m) in in year n. This value equals sf(n,m).
        This is the only method for the inflow-driven model where the lifetime distribution directly
        enters the computation.
        All other stock variables are determined by mass balance.
        The shape of the output sf array is NoofYears * NoofYears,
        and the meaning is years by age-cohorts.
        The method does nothing if the sf alreay exists.
        For example, sf could be assigned to the dynamic stock model from an exogenous computation
        to save time.
        """
        self._check_prms_set()
        for m in range(0, self._n_t):  # cohort index
            self._sf[m::, m, ...] = self._survival_by_year_id(m)

    @abstractmethod
    def _survival_by_year_id(m, **kwargs):
        pass

    @abstractmethod
    def _check_prms_set(self):
        pass

    @abstractmethod
    def set_prms(self):
        pass

    def cast_any_to_flodym_array(self, prm_in):
        if isinstance(prm_in, FlodymArray):
            prm_out = prm_in.cast_to(target_dims=self.dims).values
        else:
            prm_out = np.ndarray(self.shape)
            prm_out[...] = prm_in
        return prm_out

    def compute_outflow_pdf(self):
        """Returns an array year-by-cohort of the probability that an item
        added to stock in year m (aka cohort m) leaves in in year n. This value equals pdf(n,m).
        """
        self._sf = self.compute_survival_factor()
        pdf = np.zeros(self._shape_cohort)
        pdf[self.t_diag_indices] = 1.0 - np.moveaxis(self._sf.diagonal(0, 0, 1), -1, 0)
        for m in range(0, self._n_t):
            pdf[m + 1 :, m, ...] = -1 * np.diff(self._sf[m:, m, ...], axis=0)
        return pdf


class FixedLifetime(LifetimeModel):
    """Fixed lifetime, age-cohort leaves the stock in the model year when a certain age,
    specified as 'Mean', is reached."""

    mean: Any = None

    @model_validator(mode="after")
    def cast_mean(self):
        if self.mean is not None:
            self.mean = self.cast_any_to_flodym_array(self.mean)
        return self

    def set_prms(self, mean: FlodymArray):
        self.mean = self.cast_any_to_flodym_array(mean)

    def _check_prms_set(self):
        if self.mean is None:
            raise ValueError("Lifetime mean must be set before use.")

    def _survival_by_year_id(self, m):
        # Example: if lt is 3.5 years fixed, product will still be there after 0, 1, 2, and 3 years,
        # gone after 4 years.
        return (self._remaining_ages(m) < self.mean[m, ...]).astype(int)


class StandardDeviationLifetimeModel(LifetimeModel):
    mean: Any = None
    std: Any = None

    @model_validator(mode="after")
    def cast_mean_std(self):
        if self.mean is not None:
            self.mean = self.cast_any_to_flodym_array(self.mean)
        if self.std is not None:
            self.std = self.cast_any_to_flodym_array(self.std)
        return self

    def set_prms(self, mean: FlodymArray, std: FlodymArray):
        self.mean = self.cast_any_to_flodym_array(mean)
        self.std = self.cast_any_to_flodym_array(std)

    def _check_prms_set(self):
        if self.mean is None or self.std is None:
            raise ValueError("Lifetime mean and standard deviation must be set before use.")


class NormalLifetime(StandardDeviationLifetimeModel):
    """Normally distributed lifetime with mean and standard deviation.
    Watch out for nonzero values, for negative ages, no correction or truncation done here.
    NOTE: As normal distributions have nonzero pdf for negative ages,
    which are physically impossible, these outflow contributions can either be ignored (
    violates the mass balance) or allocated to the zeroth year of residence,
    the latter being implemented in the method compute compute_o_c_from_s_c.
    As alternative, use lognormal or folded normal distribution options.
    """

    def _survival_by_year_id(self, m):
        if np.min(self.mean) < 0:
            raise ValueError("mean must be greater than zero.")

        return scipy.stats.norm.sf(
            self._remaining_ages(m),
            loc=self.mean[m, ...],
            scale=self.std[m, ...],
        )


class FoldedNormalLifetime(StandardDeviationLifetimeModel):
    """Folded normal distribution, cf. https://en.wikipedia.org/wiki/Folded_normal_distribution
    NOTE: call this with the parameters of the normal distribution mu and sigma of curve
    BEFORE folding, curve after folding will have different mu and sigma.
    """

    def _survival_by_year_id(self, m):
        if np.min(self.mean) < 0:
            raise ValueError("mean must be greater than zero.")

        return scipy.stats.foldnorm.sf(
            self._remaining_ages(m),
            self.mean[m, ...] / self.std[m, ...],
            0,
            scale=self.std[m, ...],
        )


class LogNormalLifetime(StandardDeviationLifetimeModel):
    """Lognormal distribution
    Here, the mean and stddev of the lognormal curve, not those of the underlying normal
    distribution, need to be specified!
    Values chosen according to description on
    https://docs.scipy.org/doc/scipy-0.13.0/reference/generated/scipy.stats.lognorm.html
    Same result as EXCEL function "=LOGNORM.VERT(x;LT_LN;SG_LN;TRUE)"
    """

    def _survival_by_year_id(self, m):
        mean_square = self.mean[m, ...] * self.mean[m, ...]
        std_square = self.std[m, ...] * self.std[m, ...]
        new_mean = np.log(mean_square / np.sqrt(mean_square + std_square))
        new_std = np.sqrt(np.log(1 + std_square / mean_square))
        lt_ln = new_mean
        sg_ln = new_std
        # compute survival function
        sf_m = scipy.stats.lognorm.sf(self._remaining_ages(m), s=sg_ln, loc=0, scale=np.exp(lt_ln))
        return sf_m


class WeibullLifetime(LifetimeModel):
    """Weibull distribution with standard definition of scale and shape parameters."""

    weibull_shape: Any = None
    weibull_scale: Any = None

    @model_validator(mode="after")
    def cast_shape_scale(self):
        if self.weibull_shape is not None:
            self.weibull_shape = self.cast_any_to_flodym_array(self.weibull_shape)
        if self.weibull_scale is not None:
            self.weibull_scale = self.cast_any_to_flodym_array(self.weibull_scale)
        return self

    def set_prms(self, weibull_shape: FlodymArray, weibull_scale: FlodymArray):
        self.weibull_shape = self.cast_any_to_flodym_array(weibull_shape)
        self.weibull_scale = self.cast_any_to_flodym_array(weibull_scale)

    def _check_prms_set(self):
        if self.weibull_shape is None or self.weibull_scale is None:
            raise ValueError("Lifetime mean and standard deviation must be set before use.")

    def _survival_by_year_id(self, m):
        if np.min(self.weibull_shape) < 0:
            raise ValueError("Lifetime shape must be positive for Weibull distribution.")

        return scipy.stats.weibull_min.sf(
            self._remaining_ages(m),
            c=self.weibull_shape[m, ...],
            loc=0,
            scale=self.weibull_scale[m, ...],
        )

    # @staticmethod
    # def weibull_c_scale_from_mean_std(mean, std):
    #     """Compute Weibull parameters c and scale from mean and standard deviation.
    #     Works on scalars.
    #     Taken from https://github.com/scipy/scipy/issues/12134#issuecomment-1214031574.
    #     """
    #     def r(c, mean, std):
    #         log_mean, log_std = np.log(mean), np.log(std)
    #         # np.pi*1j is the log of -1
    #         logratio = (logsumexp([gammaln(1 + 2/c) - 2*gammaln(1+1/c), np.pi*1j])
    #                     - 2*log_std + 2*log_mean)
    #         return np.real(logratio)

    #     # Maybe a bit simpler; doesn't seem to be substantially different numerically
    #     # def r(c, mean, std):
    #     #     logratio = (gammaln(1 + 2/c) - 2*gammaln(1+1/c) -
    #     #                 logsumexp([2*log_std - 2*log_mean, 0]))
    #     #     return logratio

    #     # other methods are more efficient, but I've seen TOMS748 return garbage
    #     res = root_scalar(r, args=(mean, std), method='bisect',
    #                     bracket=[1e-300, 1e300], maxiter=2000, xtol=1e-16)
    #     assert res.converged
    #     c = res.root
    #     scale = np.exp(np.log(mean) - gammaln(1 + 1/c))
    #     return c, scale



import numpy as np
from numpy.testing import assert_array_equal
from polyfactory.factories.pydantic_factory import ModelFactory
import pytest

from flodym import FlodymArray, Dimension, DimensionSet
from flodym.flodym_array_helper import flodym_array_stack


dimension_set = DimensionSet(
    dim_list=[
        {"name": "time", "letter": "t", "items": [1990, 2000, 2010]},
        {"name": "place", "letter": "p", "items": ["World"]},
    ]
)


class FlodymArrayFactory(ModelFactory[FlodymArray]):
    dims = dimension_set
    values = np.random.rand(3, 1)


@pytest.mark.parametrize("new_dim_length", [2, 7])
def test_flodym_array_stack(new_dim_length):
    flodym_arrays = [FlodymArrayFactory.build() for _ in range(new_dim_length)]
    additional_dim = Dimension(name="extra", letter="x", items=list(range(new_dim_length)))
    stacked = flodym_array_stack(flodym_arrays, additional_dim)

    assert stacked.shape[:-1] == dimension_set.shape()
    assert stacked.dims.dim_list[:-1] == dimension_set.dim_list

    for i in range(new_dim_length):
        assert_array_equal(stacked.values[:, :, i], flodym_arrays[i].values)
    assert stacked.shape[-1] == new_dim_length
    assert stacked.dims.dim_list[-1] == additional_dim


def test_flodym_array_split():
    flodym_arrays = [FlodymArrayFactory.build() for _ in range(3)]
    items = ["pre-industrial", 1950, 2000]
    additional_dim = Dimension(name="extra", letter="x", items=items)
    stacked = flodym_array_stack(flodym_arrays, additional_dim)
    split = stacked.split(dim_letter="x")
    assert len(split) == 3
    assert list(split.keys()) == items
    for i, item in enumerate(items):
        assert_array_equal(split[item].values, flodym_arrays[i].values)



from pydantic_core import ValidationError
import pytest

from flodym import DimensionDefinition, MFADefinition, SimpleFlowDrivenStock


def test_validate_dimension_definition():
    # example that raises a validation error
    with pytest.raises(ValidationError) as error_msg:
        DimensionDefinition(name="something", letter="too_long", dtype=str)
    assert "letter" in str(error_msg.value)

    # example that does not raise an error
    DimensionDefinition(name="something", letter="s", dtype=str)


def test_validate_mfa_definition():
    # example with no dimensions defined
    with pytest.raises(ValidationError) as error_msg:
        MFADefinition(
            dimensions=[],
            processes=[],
            flows=[],
            stocks=[
                {"name": "stock_zero", "dim_letters": ("t"), "subclass": SimpleFlowDrivenStock},
            ],
            parameters=[],
        )
    assert "dim_letters" in str(error_msg.value)

    dimensions = [
        {"name": "time", "letter": "t", "dtype": int},
        {"name": "place", "letter": "p", "dtype": str},
    ]

    # example with mismatching dimensions between attributes
    with pytest.raises(ValidationError) as error_msg:
        MFADefinition(
            dimensions=dimensions,
            processes=[],
            flows=[],
            stocks=[
                {"name": "bad_stock", "dim_letters": ("t", "x"), "subclass": SimpleFlowDrivenStock},
            ],
            parameters=[],
        )
    assert "bad_stock" in str(error_msg.value)

    # example with valid dimensions in stock
    for letter_combinations in ("t", "p"), ("t",), ("p",):
        MFADefinition(
            dimensions=dimensions,
            processes=[],
            flows=[],
            stocks=[
                {
                    "name": "good_stock",
                    "dim_letters": letter_combinations,
                    "subclass": SimpleFlowDrivenStock,
                },
            ],
            parameters=[],
        )



import logging
import os
import pickle
from typing import Callable

from ..mfa_system import MFASystem
from ..flodym_arrays import FlodymArray
from .helper import to_valid_file_name


def export_mfa_to_pickle(mfa: MFASystem, export_path: str):
    """Write an MFA system to a pickle file.

    Args:
        mfa (MFASystem): The MFA system to be exported.
        export_path (str): The path to the file where the MFA system should be saved.
    """
    dict_out = convert_to_dict(mfa)
    pickle.dump(dict_out, open(export_path, "wb"))
    logging.info(f"Data saved to {export_path}")


def export_mfa_flows_to_csv(mfa: MFASystem, export_directory: str):
    """export flows of an MFA system to csv files.

    Args:
        mfa (MFASystem): The MFA system from which the flows should be exported.
        export_directory (str): The directory where the csv files should be saved.
    """
    if not os.path.exists(export_directory):
        os.makedirs(export_directory)
    for flow_name, flow in mfa.flows.items():
        path_out = os.path.join(export_directory, f"{to_valid_file_name(flow_name)}.csv")
        flow.to_df().to_csv(path_out)
    logging.info(f"Data saved in directory {export_directory}")


def export_mfa_stocks_to_csv(mfa: MFASystem, export_directory: str, with_in_and_out: bool = False):
    """export stocks of an MFA system to csv files.

    Args:
        mfa (MFASystem): The MFA system from which the stocks should be exported.
        export_directory (str): The directory where the csv files should be saved.
        with_in_and_out (bool, optional): If True, the inflow and outflow of the stocks are also exported. Defaults to False.
    """
    if not os.path.exists(export_directory):
        os.makedirs(export_directory)
    for stock_name, stock in mfa.stocks.items():
        output_items = {"stock": stock.stock}
        if with_in_and_out:
            output_items["inflow"] = stock.inflow
            output_items["outflow"] = stock.outflow
        for attribute_name, output in output_items.items():
            df = output.to_df()
            path_out = os.path.join(
                export_directory, f"{to_valid_file_name(stock_name)}_{attribute_name}.csv"
            )
            df.to_csv(path_out, index=False)
    logging.info(f"Data saved in directory {export_directory}")


def convert_to_dict(mfa: MFASystem, type: str = "numpy") -> dict:
    """Convert an MFA system to a dictionary which is readable without flodym.

    Args:
        mfa (MFASystem): The MFA system to be converted.
        type (str, optional): The type of the values in the flows and stocks. Options are 'numpy' and 'pandas'. Defaults to "numpy".

    Returns:
        dict: The MFA system as a dictionary. Contains the items 'dimension_names', 'dimension_items', 'processes', 'flows', 'flow_dimensions', 'flow_processes', 'stocks', 'stock_dimensions', 'stock_processes'.
    """
    convert_func = _get_convert_func(type)
    return _convert_to_dict_by_func(mfa, convert_func)


def _get_convert_func(type: str):
    if type == "numpy":

        def convert_func(array: FlodymArray):
            return array.values

    elif type == "pandas":

        def convert_func(array: FlodymArray):
            return array.to_df()

    else:
        raise ValueError(f"Unknown export type {type}. Must be 'numpy' or 'pandas'.")
    return convert_func


def _convert_to_dict_by_func(mfa: MFASystem, convert_func: Callable) -> dict:
    dict_out = {}
    dict_out["dimension_names"] = {d.letter: d.name for d in mfa.dims}
    dict_out["dimension_items"] = {d.name: d.items for d in mfa.dims}
    dict_out["processes"] = [p.name for p in mfa.processes.values()]
    dict_out["flows"] = {n: convert_func(f) for n, f in mfa.flows.items()}
    dict_out["flow_dimensions"] = {n: f.dims.letters for n, f in mfa.flows.items()}
    dict_out["flow_processes"] = {
        n: (f.from_process.name, f.to_process.name) for n, f in mfa.flows.items()
    }
    dict_out["stocks"] = {s_name: convert_func(s.stock) for s_name, s in mfa.stocks.items()}
    dict_out["stock_dimensions"] = {
        s_name: s.stock.dims.letters for s_name, s in mfa.stocks.items()
    }
    dict_out["stock_processes"] = {
        s_name: s.process.name for s_name, s in mfa.stocks.items() if s.process is not None
    }
    return dict_out



import logging
import numpy as np
import pandas as pd
from typing import Literal, Optional, TYPE_CHECKING, Iterable
from pydantic import BaseModel as PydanticBaseModel
import itertools

from .dimensions import Dimension

if TYPE_CHECKING:
    from .flodym_arrays import FlodymArray


class FlodymDataFormat(PydanticBaseModel):

    type: Literal["long", "wide"]
    value_column: str = "value"
    columns_dim: Optional[str] = None


class DataFrameToFlodymDataConverter:
    """Converts a panda DataFrame with various possible formats to a numpy array that can be used
    as values of a FlodymArray.

    Usually not called by the user, but from within the FlodymArray from_df and
    set_values_from_df methods.

    In case of errors, turning on debug logging might help to understand the process.
    """

    def __init__(self, df: pd.DataFrame, flodym_array: "FlodymArray"):
        self.df = df.copy()
        self.flodym_array = flodym_array
        self.target_values = self.get_target_values()

    def get_target_values(self) -> np.ndarray:
        logging.debug(
            f"Start setting values for FlodymArray {self.flodym_array.name} with dimensions {self.flodym_array.dims.names} from dataframe."
        )
        self._reset_non_default_index()
        self._determine_format()
        self._df_to_long_format()
        self._check_missing_dim_columns()
        self._convert_type()
        self._sort_df()
        self._check_data_complete()
        return self.df[self.format.value_column].values.reshape(self.flodym_array.shape)

    def _reset_non_default_index(self):
        if isinstance(self.df.index, pd.MultiIndex):
            self.df.reset_index(inplace=True)
        elif self.df.index.name is not None:
            self.df.reset_index(inplace=True)
        elif self.df.index.dtype != np.int64:
            self.df.reset_index(inplace=True)
        elif self.df.index.min() >= 1700 and self.df.index.max() <= 2300:
            self.df.reset_index(inplace=True)

    def _determine_format(self):
        self._get_dim_columns_by_name()
        self._check_if_first_row_are_items()
        self._check_for_dim_columns_by_items()
        self._check_value_columns()

    def _get_dim_columns_by_name(self):
        self.dim_columns = [c for c in self.df.columns if c in self.flodym_array.dims.names]
        logging.debug(f"Recognized index columns by name: {self.dim_columns}")

    def _check_if_first_row_are_items(self):
        """If data without columns names was read, but the first row was assumed to be column names,
        the first row of the data frame might erroneously end up as column names.
        This method checks if that is the case, and if so, prepends a row based on column names.
        """
        column_name = self.df.columns[0]
        col_items = self.df[column_name].unique()
        extended_col_items = [column_name] + col_items.tolist()
        for dim in self.flodym_array.dims:
            if self.same_items(extended_col_items, dim):
                self._add_column_names_as_row(column_name, dim)

    def _add_column_names_as_row(self, column_name: str, dim: Dimension):
        if len(self.dim_columns) > 0:
            raise ValueError(
                f"Ambiguouity detected: column with first item {column_name} could be "
                f"dimension {dim.name} if the first row counts as an item, but columns "
                f"{self.dim_columns} are already recognized as dimensions first row as name."
                f" Please change the item names of the affected dimension, or use a"
                f" different method to read data."
            )
        # prepend a row to df with all the column names
        self.df = pd.concat([pd.DataFrame([self.df.columns], columns=self.df.columns), self.df])
        # rename columns to range from 0 to n, save original name
        column_name = f"column {self.df.columns.get_loc(column_name)}"
        self.df.columns = [f"column {i}" for i in range(len(self.df.columns))]

    def _check_for_dim_columns_by_items(self):
        for cn in self.df.columns:
            if cn in self.dim_columns:
                continue
            found = self._check_if_dim_column_by_items(cn)
            if not found:
                logging.debug(
                    f"Could not find dimension with same items as column {cn}. "
                    "Assuming this is the first value column; Won't look further."
                )
                return

    def _check_if_dim_column_by_items(self, column_name: str) -> bool:
        logging.debug(f"Checking if {column_name} is a dimension by comparing items with dim items")
        col_items = self.df[column_name].unique()
        for dim in self.flodym_array.dims:
            if self.same_items(col_items, dim):
                logging.debug(f"{column_name} is dimension {dim.name}.")
                self.df.rename(columns={column_name: dim.name}, inplace=True)
                self.dim_columns.append(dim.name)
                return True
        return False

    def _check_value_columns(self):
        value_cols = np.setdiff1d(list(self.df.columns), self.dim_columns)
        logging.debug(f"Assumed value columns: {value_cols}")
        value_cols_are_dim_items = self._check_if_value_columns_match_dim_items(value_cols)
        if not value_cols_are_dim_items:
            self._check_if_valid_long_format(value_cols)

    def _check_if_value_columns_match_dim_items(self, value_cols: list[str]) -> bool:
        logging.debug("Trying to match set of value column names with items of dimension.")
        for dim in self.flodym_array.dims:
            if self.same_items(value_cols, dim):
                logging.debug(f"Value columns match dimension items of {dim.name}.")
                self.format = FlodymDataFormat(type="wide", columns_dim=dim.name)
                if dim.dtype is not None:
                    for c in value_cols:
                        self.df.rename(columns={c: dim.dtype(c)}, inplace=True)
                return True
        return False

    def _check_if_valid_long_format(self, value_cols: list[str]):
        logging.debug(
            "Could not find dimension with same item set as value column names. Assuming long format, i.e. one value column."
        )
        if len(value_cols) == 1:
            self.format = FlodymDataFormat(type="long", value_column=value_cols[0])
            logging.debug(f"Value column name is {value_cols[0]}.")
        else:
            raise ValueError(
                "More than one value columns. Could not find a dimension the items of which match the set of value column names. "
                f"Value columns: {value_cols}. Please check input data for format, typos, data types and missing items."
            )

    def _df_to_long_format(self):
        if self.format.type != "wide":
            return
        logging.debug("Converting wide format to long format.")
        value_cols = self.flodym_array.dims[self.format.columns_dim].items
        self.df = self.df.melt(
            id_vars=[c for c in self.df.columns if c not in value_cols],
            value_vars=value_cols,
            var_name=self.format.columns_dim,
            value_name=self.format.value_column,
        )
        self.dim_columns.append(self.format.columns_dim)
        self.format = FlodymDataFormat(type="long", value_column=self.format.value_column)

    def _check_missing_dim_columns(self):
        missing_dim_columns = np.setdiff1d(list(self.flodym_array.dims.names), self.dim_columns)
        for c in missing_dim_columns:
            if len(self.flodym_array.dims[c].items) == 1:
                self.df[c] = self.flodym_array.dims[c].items[0]
                self.dim_columns.append(c)
            else:
                raise ValueError(
                    f"Dimension {c} from array has more than one item, but is not found in df. Please specify column in dataframe."
                )

    def _convert_type(self):
        for dim in self.flodym_array.dims:
            if dim.dtype is not None:
                self.df[dim.name] = self.df[dim.name].map(dim.dtype)
        self.df[self.format.value_column] = self.df[self.format.value_column].astype(np.float64)

    def _sort_df(self):
        """Sort the columns of the data frame according to the order of the dimensions in the
        FlodymArray.
        Sort the rows of the data frame according to the order of the dimension items in the
        FlodymArray.
        """
        # sort columns
        self.df = self.df[list(self.flodym_array.dims.names) + [self.format.value_column]]
        # sort rows
        self.df = self.df.sort_values(
            by=list(self.flodym_array.dims.names),
            key=lambda x: x.map(lambda y: self.flodym_array.dims[x.name].items.index(y)),
        )

    def _check_data_complete(self):
        # Generate expected index tuples from FlodymArray dimensions
        if self.flodym_array.dims.ndim == 0:
            expected_index_tuples = set()
        else:
            expected_index_tuples = set(
                itertools.product(*(dim.items for dim in self.flodym_array.dims))
            )

        # Generate actual index tuples from DataFrame columns
        actual_index_tuples = set(
            self.df.drop(columns=self.format.value_column).itertuples(index=False, name=None)
        )

        # Compare the two sets
        if expected_index_tuples != actual_index_tuples:
            missing_items = expected_index_tuples - actual_index_tuples
            unexpected_items = actual_index_tuples - expected_index_tuples
            raise ValueError(
                f"Dataframe index mismatch! Missing items: {missing_items}, Unexpected items: {unexpected_items}"
            )
        if any(self.df[self.format.value_column].isna()):
            raise ValueError("Empty cells/NaN values in value column!")

    @staticmethod
    def same_items(arr: Iterable, dim: Dimension) -> bool:
        if dim.dtype is not None:
            try:
                arr = [dim.dtype(a) for a in arr]
            except ValueError:
                return False
        return len(set(arr).symmetric_difference(set(dim.items))) == 0



from pydantic import BaseModel as PydanticBaseModel, model_validator, ConfigDict, computed_field
from typing import Optional, Any
import numpy as np
import plotly.graph_objects as go
import plotly as pl

from ..mfa_system import MFASystem
from ..flodym_arrays import Flow
from .helper import CustomNameDisplayer


class PlotlySankeyPlotter(CustomNameDisplayer, PydanticBaseModel):

    model_config = ConfigDict(arbitrary_types_allowed=True, protected_namespaces=())

    mfa: MFASystem
    """MFA system to visualize."""
    slice_dict: Optional[dict] = {}
    """for selection of a subset of the data; all other dimensions are summed over"""
    node_color_dict: Optional[dict] = {"default": "gray"}
    """color of the nodes (processes and stocks)"""
    flow_color_dict: Optional[dict] = {"default": "hsl(230,20,70)"}
    """dictionary of colors for flows.
    Keys are flow names, values are either a single color or a tuple of the dimension names
    to split the flow by, and a color scheme as a list of colors. There must be a "default" key
    to resort to if a flow is not in the dictionary.
    """
    exclude_processes: Optional[list[str]] = ["sysenv"]
    """processes that won't show up in the plot; neither will flows to and from them"""
    exclude_flows: Optional[list[str]] = []
    """flows that won't show up in the plot"""

    @model_validator(mode="after")
    def check_dims(self):
        for dim_letter in self.slice_dict.keys():
            if dim_letter not in self.mfa.dims.letters:
                raise ValueError(f"Dimension {dim_letter} given in slice_dict not in DimensionSet.")
        return self

    @model_validator(mode="after")
    def check_excluded(self):
        for p in self.exclude_processes:
            if p not in self.mfa.processes:
                raise ValueError(f"Process {p} given in exclude_processes not in MFASystem.")
        for f in self.exclude_flows:
            if f not in self.mfa.flows:
                raise ValueError(f"Flow {f} given in exclude_flows not in MFASystem.")
        return self

    @model_validator(mode="after")
    def check_flow_colors(self):
        for f in self.shown_flows.values():
            if "default" not in self.flow_color_dict:
                raise ValueError(
                    f"flow_color_dict must have a 'default' key to resort to if a flow is not in the dictionary"
                )
            if f.name not in self.flow_color_dict:
                self.flow_color_dict[f.name] = self.flow_color_dict["default"]
                fallback_str = " (not found in dict, using default color)"
            else:
                fallback_str = ""
            self._check_flow_color(f, fallback_str)
        return self

    @model_validator(mode="after")
    def check_node_colors(self):
        for p in self.shown_processes:
            if "default" not in self.node_color_dict:
                raise ValueError(
                    f"node_color_dict must have a 'default' key to resort to if a process is not in the dictionary"
                )
            if p.name not in self.node_color_dict:
                self.node_color_dict[p.name] = self.node_color_dict["default"]
                fallback_str = " (not found in dict, using default color)"
            else:
                fallback_str = ""
            if not isinstance(self.node_color_dict[p.name], str):
                raise ValueError(
                    f"Color for process {p.name}{fallback_str} must be a string, not a {type(self.node_color_dict[p.name])}."
                )
        return self

    def _check_flow_color(self, f: Flow, fallback_str):
        color = self.flow_color_dict[f.name]
        if isinstance(color, str):
            return
        elif not isinstance(color, tuple):
            raise ValueError(
                f"In flow_color_dict, the value for flow {f.name}{fallback_str} must be either a color string or a tuple of dimension name and color list"
            )

        if len(color) != 2:
            raise ValueError(
                f"In flow_color_dict, color tuple for flow {f.name}{fallback_str} must have length 2."
            )
        if color[0] not in f.dims:
            raise ValueError(
                f"In flow_color_dict, first element of color tuple for flow {f.name}{fallback_str} must be a dimension in flow {f.name}"
            )
        if not isinstance(color[1], list):
            raise ValueError(
                f"In flow_color_dict, second element of color tuple for flow {f.name}{fallback_str} must be a list of colors"
            )
        if len(color[1]) < self.mfa.dims[color[0]].len:
            raise ValueError(
                f"In flow_color_dict, list in second element of color tuple for flow {f.name}{fallback_str} must not be shorter than dimension {color[0]}"
            )

    @property
    def shown_processes(self):
        return [p for p in self.mfa.processes.values() if p.name not in self.exclude_processes]

    @property
    def excluded_process_ids(self):
        return [p.id for p in self.mfa.processes.values() if p.name in self.exclude_processes]

    @property
    def shown_flows(self) -> dict[str, Flow]:
        return {f.name: f for f in self.mfa.flows.values() if self._flow_is_shown(f)}

    def _flow_is_shown(self, f: Flow):
        return not (
            (f.name in self.exclude_flows)
            or (f.from_process_id in self.excluded_process_ids)
            or (f.to_process_id in self.excluded_process_ids)
        )

    @property
    def ids_in_sankey(self):
        return {p.id: i for i, p in enumerate(self.shown_processes)}

    def plot(self):
        links = self._get_links_dict()
        nodes = self._get_nodes_dict()
        return self._get_fig(links, nodes)

    def _get_links_dict(self):
        links = DictOfLists(list_names=["source", "target", "value", "label", "color"])
        for f in self.shown_flows.values():
            self._append_flow(f, links)
        return links.dict

    def _append_flow(self, f: Flow, links: "DictOfLists"):
        source = self.ids_in_sankey[f.from_process.id]
        target = self.ids_in_sankey[f.to_process.id]
        label = self.display_name(f.name)

        slice_dict = {k: v for k, v in self.slice_dict.items() if k in f.dims.letters}
        f_slice = f[slice_dict]
        color = self.flow_color_dict[f.name]
        if isinstance(color, tuple):
            split_flows_by = color[0]
            colors = color[1]
            splitting_letter_tuple = (self.mfa.dims[split_flows_by].letter,)
            values = f_slice.sum_values_to(splitting_letter_tuple)
            for v, c in zip(values, colors):
                links.append(label=label, source=source, target=target, color=c, value=v)
        else:
            links.append(
                label=label,
                source=source,
                target=target,
                color=color,
                value=f_slice.sum_values(),
            )

    def _get_nodes_dict(self):
        return {
            "label": [self.display_name(p.name) for p in self.shown_processes],
            "color": [self.node_color_dict[p.name] for p in self.shown_processes],
            "pad": 10,
        }

    def _get_fig(self, links: dict, nodes: dict) -> go.Figure:
        fig = go.Figure(
            go.Sankey(
                arrangement="snap",
                link=links,
                node=nodes,
            )
        )
        return fig


class DictOfLists:

    def __init__(self, list_names: list[str]):
        self.dict = {l: [] for l in list_names}

    def append(self, **kwargs):
        for k, v in kwargs.items():
            self.dict[k].append(v)



"""Home to various `Stock` classes,
including flow-driven stocks and dynamic (lifetime-based) stocks.
"""

from abc import abstractmethod
import numpy as np
from pydantic import BaseModel as PydanticBaseModel, ConfigDict, model_validator
from typing import Optional, Union
import logging

from .processes import Process
from .flodym_arrays import StockArray
from .dimensions import DimensionSet
from .lifetime_models import LifetimeModel


class Stock(PydanticBaseModel):
    """Stock objects are components of an MFASystem, where materials can accumulate over time.
    They consist of three :py:class:`flodym.FlodymArray` objects:
    stock (the accumulation), inflow, outflow.

    The base class only allows to compute the stock from known inflow and outflow.
    The subclasses allows computations using a lifetime distribution function,
    which is necessary if not both inflow and outflow are known.
    """

    model_config = ConfigDict(protected_namespaces=(), arbitrary_types_allowed=True)

    dims: DimensionSet
    """Dimensions of the stock, inflow, and outflow arrays. Time must be the first dimension."""
    stock: Optional[StockArray] = None
    """Accumulation of the stock"""
    inflow: Optional[StockArray] = None
    """Inflow into the stock"""
    outflow: Optional[StockArray] = None
    """Outflow from the stock"""
    name: Optional[str] = "unnamed"
    """Name of the stock"""
    process: Optional[Process] = None
    """Process the stock is associated with, if any. Needed for example for the mass balance."""
    time_letter: str = "t"
    """Letter of the time dimension in the dimensions set, to make sure it's the first one."""

    @model_validator(mode="after")
    def validate_stock_arrays(self):
        if self.stock is None:
            self.stock = StockArray(dims=self.dims, name=f"{self.name}_stock")
        elif self.stock.dims.letters != self.dims.letters:
            raise ValueError(
                f"Stock dimensions {self.stock.dims.letters} do not match prescribed dims {self.dims.letters}."
            )
        if self.inflow is None:
            self.inflow = StockArray(dims=self.dims, name=f"{self.name}_inflow")
        elif self.inflow.dims.letters != self.dims.letters:
            raise ValueError(
                f"Inflow dimensions {self.inflow.dims.letters} do not match prescribed dims {self.dims.letters}."
            )
        if self.outflow is None:
            self.outflow = StockArray(dims=self.dims, name=f"{self.name}_outflow")
        elif self.outflow.dims.letters != self.dims.letters:
            raise ValueError(
                f"Outflow dimensions {self.outflow.dims.letters} do not match prescribed dims {self.dims.letters}."
            )
        return self

    @model_validator(mode="after")
    def validate_time_first_dim(self):
        if self.dims.letters[0] != self.time_letter:
            raise ValueError(
                f"Time dimension must be the first dimension, i.e. time_letter (now {self.time_letter}) must be the first letter in dims.letters (now {self.dims.letters[0]})."
            )
        return self

    @abstractmethod
    def compute(self):
        # always add this check first
        self._check_needed_arrays()

    @abstractmethod
    def _check_needed_arrays(self):
        pass

    @property
    def shape(self) -> tuple:
        """Shape of the stock, inflow, outflow arrays, defined by the dimensions."""
        return self.dims.shape()

    @property
    def process_id(self) -> int:
        """ID of the process the stock is associated with."""
        return self.process.id

    def to_stock_type(self, desired_stock_type: type, **kwargs):
        """Return an object of a new stock type with values and dimensions the same as the original.
        `**kwargs` can be used to pass additional model attributes as required by the desired stock
        type, if these are not contained in the original stock type.
        """
        return desired_stock_type(**self.__dict__, **kwargs)

    def check_stock_balance(self):
        balance = self.get_stock_balance()
        balance = np.max(np.abs(balance).sum(axis=0))
        if balance > 1:  # 1 tonne accuracy
            raise RuntimeError("Stock balance for dynamic stock model is too high: " + str(balance))
        elif balance > 0.001:
            print("Stock balance for model dynamic stock model is noteworthy: " + str(balance))

    def get_stock_balance(self) -> np.ndarray:
        """Check whether inflow, outflow, and stock are balanced.
        If possible, the method returns the vector 'Balance', where Balance = inflow - outflow - stock_change
        """
        dsdt = np.diff(
            self.stock.values, axis=0, prepend=0
        )  # stock_change(t) = stock(t) - stock(t-1)
        return self.inflow.values - self.outflow.values - dsdt


class SimpleFlowDrivenStock(Stock):
    """Given inflows and outflows, the stock can be calculated."""

    def _check_needed_arrays(self):
        if (
            np.max(np.abs(self.inflow.values)) < 1e-10
            and np.max(np.abs(self.outflow.values)) < 1e-10
        ):
            logging.warning("Inflow and Outflow are zero. This will lead to a zero stock.")

    def compute(self):
        self._check_needed_arrays()
        self.stock.values[...] = np.cumsum(self.inflow.values - self.outflow.values, axis=0)


class DynamicStockModel(Stock):
    """Parent class for dynamic stock models, which are based on stocks having a specified
    lifetime (distribution).
    """

    lifetime_model: Union[LifetimeModel, type]
    """Lifetime model, which contains the lifetime distribution function.
    Can be input either as a LifetimeModel subclass, or as an instance of a
    LifetimeModel subclass. For available subclasses, see `flodym.lifetime_models`.
    """
    _outflow_by_cohort: np.ndarray = None
    _stock_by_cohort: np.ndarray = None

    @model_validator(mode="after")
    def init_cohort_arrays(self):
        self._stock_by_cohort = np.zeros(self._shape_cohort)
        self._outflow_by_cohort = np.zeros(self._shape_cohort)
        return self

    @model_validator(mode="after")
    def init_lifetime_model(self):
        if isinstance(self.lifetime_model, type):
            if not issubclass(self.lifetime_model, LifetimeModel):
                raise ValueError("lifetime_model must be a subclass of LifetimeModel.")
            self.lifetime_model = self.lifetime_model(dims=self.dims, time_letter=self.time_letter)
        elif self.lifetime_model.dims.letters != self.dims.letters:
            raise ValueError("Lifetime model dimensions do not match stock dimensions.")
        return self

    def _check_needed_arrays(self):
        self.lifetime_model._check_prms_set()

    @property
    def _n_t(self) -> int:
        return list(self.shape)[0]

    @property
    def _shape_cohort(self) -> tuple:
        return (self._n_t,) + self.shape

    @property
    def _shape_no_t(self) -> tuple:
        return tuple(list(self.shape)[1:])

    @property
    def _t_diag_indices(self) -> tuple:
        return np.diag_indices(self._n_t) + (slice(None),) * len(self._shape_no_t)


class InflowDrivenDSM(DynamicStockModel):
    """Inflow driven model.
    Given inflow and lifetime distribution calculate stocks and outflows.
    """

    def _check_needed_arrays(self):
        super()._check_needed_arrays()
        if np.allclose(self.inflow.values, np.zeros(self.shape)):
            logging.warning("Inflow is zero. This will lead to a zero stock and outflow.")

    def compute(self):
        """Determine stocks and outflows and store values in the class instance."""
        self._check_needed_arrays()
        self.compute_stock_by_cohort()
        self.compute_outflow_by_cohort()
        self.stock.values[...] = self._stock_by_cohort.sum(axis=1)
        self.outflow.values[...] = self._outflow_by_cohort.sum(axis=1)

    def compute_stock_by_cohort(self) -> np.ndarray:
        """With given inflow and lifetime distribution, the method builds the stock by cohort.
        s_c[t,c] = i[c] * sf[t,c] for all t, c
        from the perspective of the stock the inflow has the dimension age-cohort,
        as each inflow(t) is added to the age-cohort c = t
        """
        self._stock_by_cohort = np.einsum(
            "c...,tc...->tc...", self.inflow.values, self.lifetime_model.sf
        )

    def compute_outflow_by_cohort(self) -> np.ndarray:
        """Compute outflow by cohort from changes in the stock by cohort and the known inflow."""
        self._outflow_by_cohort[1:, :, ...] = -np.diff(self._stock_by_cohort, axis=0)
        # allow for outflow in year 0 already
        self._outflow_by_cohort[self._t_diag_indices] = self.inflow.values - np.moveaxis(
            self._stock_by_cohort.diagonal(0, 0, 1), -1, 0
        )


class StockDrivenDSM(DynamicStockModel):
    """Stock driven model.
    Given total stock and lifetime distribution, calculate inflows and outflows.
    """

    def _check_needed_arrays(self):
        super()._check_needed_arrays()
        if np.allclose(self.stock.values, np.zeros(self.shape)):
            logging.warning("Stock is zero. This will lead to a zero inflow and outflow.")

    def compute(self):
        """Determine inflows and outflows and store values in the class instance."""
        self._check_needed_arrays()
        self.compute_inflow_and_outflow()
        self.outflow.values[...] = self._outflow_by_cohort.sum(axis=1)

    def compute_inflow_and_outflow(self) -> tuple[np.ndarray]:
        """With given total stock and lifetime distribution,
        the method builds the stock by cohort and the inflow."""
        sf = self.lifetime_model.sf
        # construct the sf of a product of cohort tc remaining in the stock in year t
        # First year:
        self.inflow.values[0, ...] = np.where(
            sf[0, 0, ...] != 0.0, self.stock.values[0] / sf[0, 0], 0.0
        )
        # Future decay of age-cohort of year 0.
        self._stock_by_cohort[:, 0, ...] = self.inflow.values[0, ...] * sf[:, 0, ...]
        self._outflow_by_cohort[0, 0, ...] = (
            self.inflow.values[0, ...] - self._stock_by_cohort[0, 0, ...]
        )
        # all other years:
        for m in range(1, self._n_t):  # for all years m, starting in second year
            # 1) Compute outflow from previous age-cohorts up to m-1
            # outflow table is filled row-wise, for each year m.
            self._outflow_by_cohort[m, 0:m, ...] = (
                self._stock_by_cohort[m - 1, 0:m, ...] - self._stock_by_cohort[m, 0:m, ...]
            )
            self.inflow_from_balance(m)

    def inflow_from_balance(self, m: int) -> np.ndarray:
        """determine inflow from mass balance and do not correct negative inflow"""

        sf = self.lifetime_model.sf
        # allow for outflow during first year by rescaling with 1/sf[m,m]
        self.inflow.values[m, ...] = np.where(
            sf[m, m, ...] != 0.0,
            (self.stock.values[m, ...] - self._stock_by_cohort[m, :, ...].sum(axis=0))
            / sf[m, m, ...],
            0.0,
        )
        # 3) Add new inflow to stock and determine future decay of new age-cohort
        self._stock_by_cohort[m::, m, ...] = self.inflow.values[m, ...] * sf[m::, m, ...]
        self._outflow_by_cohort[m, m, ...] = self.inflow.values[m, ...] * (1 - sf[m, m, ...])


class StockDrivenDSM_NIC(StockDrivenDSM):

    def inflow_from_balance(self, m):
        is_negative_inflow = self.check_negative_inflow(m)
        if is_negative_inflow:
            self.inflow_from_balance_correction(m)
        else:
            super().inflow_from_balance(m)

    def check_negative_inflow(self, m: int) -> bool:
        """Check if inflow is negative."""
        inflow_test = self.stock.values[m, ...] - self._stock_by_cohort[m, :, ...].sum(axis=0)
        return inflow_test < 0

    def inflow_from_balance_correction(self, m: int) -> np.ndarray:
        """determine inflow from mass balance and correct negative inflow

        NOTE: This method of negative inflow correction is only one of many plausible methods of increasing the
        outflow to keep matching stock levels. It assumes that the surplus stock is removed in the year that
        it becomes obsolete. Each cohort loses the same fraction. Modellers need to try out whether this
        method leads to justifiable results. In some situations it is better to change the lifetime assumption
        than using the NegativeInflowCorrect option.
        """

        # if the stock declines faster than according to the lifetime model, this option allows to extract
        # additional stock items.
        # The negative inflow correction implemented here was developed in a joined effort by Sebastiaan Deetman
        # and Stefan Pauliuk.

        # delta = -inflow
        delta = self._stock_by_cohort[m, :, ...].sum(axis=0) - self.stock.values[m, ...]
        # Set inflow to 0 and distribute mass balance gap onto remaining cohorts:
        self.inflow.values[m, ...] = 0
        delta_percent = np.where(
            self._stock_by_cohort[m, :, ...].sum(axis=0) != 0,
            delta / self._stock_by_cohort[m, :, ...].sum(axis=0),
            0.0,
        )
        # - Distribute gap equally across all cohorts (each cohort is adjusted by the same %, based on
        #   surplus with regards to the prescribed stock)
        # - delta_percent is a % value <= 100%
        # - correct for outflow and stock in current and future years
        # - adjust the entire stock AFTER year m as well, stock is lowered in year m, so future cohort
        #   survival also needs to decrease.

        # increase outflow according to the lost fraction of the stock, based on Delta_c
        self._outflow_by_cohort[m, :, ...] = self._outflow_by_cohort[m, :, ...] + (
            self._stock_by_cohort[m, :, ...] * delta_percent
        )
        # shrink future description of stock from previous age-cohorts by factor Delta_percent in current
        # AND future years.
        self._stock_by_cohort[m::, 0:m, ...] = self._stock_by_cohort[m::, 0:m, ...] * (
            1 - delta_percent
        )



"""Home to helper functions for the `Stock` class."""

from .processes import Process
from .flodym_array_helper import flodym_array_stack
from .dimensions import Dimension, DimensionSet
from .mfa_definition import StockDefinition
from .stocks import Stock


def stock_stack(stocks: list[Stock], dimension: Dimension) -> Stock:
    """Make a `FlowDrivenStock` object as a combination of `Stock` objects,
    by combining them on a new dimension.
    For example, we could have one stock for each material of interest, and
    with this function we can combine them to a stock object that contains
    information about all the materials.
    """
    stacked_stock = flodym_array_stack([stock.stock for stock in stocks], dimension=dimension)
    stacked_inflow = flodym_array_stack([stock.inflow for stock in stocks], dimension=dimension)
    stacked_outflow = flodym_array_stack([stock.outflow for stock in stocks], dimension=dimension)
    return stocks[0].__class__(
        stock=stacked_stock,
        inflow=stacked_inflow,
        outflow=stacked_outflow,
        name=stocks[0].name,
        process=stocks[0].process,
    )


def make_empty_stocks(
    stock_definitions: list[StockDefinition],
    processes: dict[str, Process],
    dims: DimensionSet,
) -> dict[str, Stock]:
    """Initialise empty Stock objects for each of the stocks listed in stock definitions."""
    empty_stocks = {}
    for stock_definition in stock_definitions:
        dim_subset = dims.get_subset(stock_definition.dim_letters)
        if stock_definition.process_name is None:
            process = None
        else:
            try:
                process = processes[stock_definition.process_name]
            except KeyError:
                raise KeyError(f"Process {stock_definition.process_name} not in processes.")

        init_args = dict(
            dims=dim_subset,
            time_letter=stock_definition.time_letter,
            name=stock_definition.name,
            process=process,
        )
        if stock_definition.lifetime_model_class is not None:
            lifetime_model = stock_definition.lifetime_model_class(
                dims=dim_subset, time_letter=stock_definition.time_letter
            )
            init_args["lifetime_model"] = lifetime_model

        stock = stock_definition.subclass(**init_args)
        empty_stocks[stock.name] = stock
    return empty_stocks



"""The classes and methods defined here are building blocks for creating MFA systems.
This includes the base `FlodymArray` class and its helper the `SubArrayHandler`,
as well as applications of the `FlodymArray` for specific model components.
"""

from collections.abc import Iterable
from copy import deepcopy
from collections import defaultdict
import numpy as np
import pandas as pd
from pydantic import BaseModel as PydanticBaseModel, ConfigDict, model_validator
from typing import Optional, Union

from .processes import Process
from .dimensions import DimensionSet, Dimension
from ._df_to_flodym_array import DataFrameToFlodymDataConverter


def _is_iterable(arg):
    return isinstance(arg, Iterable) and not isinstance(arg, (str, Dimension))


class FlodymArray(PydanticBaseModel):
    """Parent class for an array with pre-defined dimensions, which are addressed by name. Operations between
    different multi-dimensional arrays can than be performed conveniently, as the dimensions are automatically matched.

    In order to 'fix' the dimensions of the array, the array has to be 'declared' by calling the FlodymArray object
    constructor with a set of dimensions before working with it.
    Basic mathematical operations between FlodymArrays are defined, which return a FlodymArray object as a result.

    In order to set the values of a FlodymArray object to that of another one, the ellipsis slice ('[...]') can be
    used, e.g.
    foo[...] = bar.
    This ensures that the dimensionality of the array (foo) is not changed, and that the dimensionality of the
    right-hand side FlodymArray (bar) is consistent.
    While the syntaxes like of 'foo = bar + baz' are also possible (where 'bar' and 'baz' are FlodymArrays),
    it is not recommended, as it provides no control over the dimensionality of 'foo'. Use foo[...] = bar + baz instead.

    The values of the FlodymArray object are stored in a numpy array, and can be accessed directly via the 'values'
    attribute.
    So if type(bar) is np.ndarray, the operation
    foo.values[...] = bar
    is also possible.
    It is not recommended to use 'foo.values = bar' without the slice, as this might change the dimensionality of
    foo.values.

    Subsets of arrays can be set or retrieved.
    Here, slicing information is passed instead of the ellipsis to the square brackets of the FlodymArray, i.e.
    foo[keys] = bar or foo = bar[keys]. For details on the allowed values of 'keys', see the docstring of the
    SubArrayHandler class.

    The dimensions of a FlodymArray stored as a :py:class:`flodym.DimensionSet` object in the 'dims' attribute.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, protected_namespaces=())

    dims: DimensionSet
    """Dimensions of the FlodymArray."""
    values: Optional[Union[np.ndarray, np.generic]] = None
    """Values of the FlodymArray. Must have the same shape as the dimensions of the FlodymArray. If None, an array of zeros is created."""
    name: Optional[str] = "unnamed"
    """Name of the FlodymArray."""

    @model_validator(mode="after")
    def validate_values(self):
        if self.values is None:
            self.values = np.zeros(self.dims.shape())
        self._check_value_format()
        return self

    def _check_value_format(self):
        if not isinstance(self.values, (np.ndarray, np.generic)):
            raise ValueError("Values must be a numpy array or numpy generic.")
        if self.dims.ndim > 0 and not isinstance(self.values, np.ndarray):
            raise ValueError("Values must be a numpy array, except for 0-dimensional arrays.")
        elif self.dims.ndim == 0 and isinstance(self.values, np.generic):
            self.values = np.array(self.values)

        if self.values.shape != self.dims.shape():
            raise ValueError(
                f"Values passed to {self.__class__.__name__} must have the same shape as the DimensionSet.\n"
                f"Array shape: {self.dims.shape()}\n"
                f"Values shape: {self.values.shape}\n"
            )

    @classmethod
    def from_dims_superset(
        cls, dims_superset: DimensionSet, dim_letters: tuple = None, **kwargs
    ) -> "FlodymArray":
        """Create a FlodymArray object from a superset of dimensions, by specifying which
        dimensions to take.

        Parameters:
            dims_superset: DimensionSet from which the objects dimensions are derived
            dim_letters: specify which dimensions to take from dims_superset
            kwargs: additional keyword arguments passed to the FlodymArray constructor

        Returns:
            cls instance
        """
        dims = dims_superset.get_subset(dim_letters)
        return cls(dims=dims, **kwargs)

    @classmethod
    def from_df(cls, dims: DimensionSet, df: pd.DataFrame, **kwargs) -> "FlodymArray":
        """Create a FlodymArray object from a DataFrame.

        Parameters:
            dims (DimensionSet): Dimensions of the FlodymArray
            df (DataFrame): pandas DataFrame containing the values of the FlodymArray.
                Dimensions of the named dim array can be given in DataFrame columns or the index.
                The DataFrame can be in long or wide format, that is there can either be one value column,
                or the value columns are named by items of one FlodymArray dimension.
                If dimension names are not given in the respective index or column, they are inferred from the
                items of the dimensions of the FlodymArray.
                It is advisable to give the dimension names in the DataFrame, as this makes the error messages
                more informative if there are typos in the items or if items are missing.
                Ordering of rows and columns is arbitrary, but the items across each dimension must be given,
                must be complete and exactly match those of the FlodymArray.
                Dimensions with only one item do not need to be given in the DataFrame.
                Supersets of dimensions (i.e. additional values) will lead to an error.

        Returns:
            FlodymArray: FlodymArray object with the values from the DataFrame
        """
        flodym_array = cls(dims=dims, **kwargs)
        flodym_array.set_values_from_df(df)
        return flodym_array

    def _sub_array_handler(self, definition) -> "SubArrayHandler":
        return SubArrayHandler(self, definition)

    @property
    def shape(self) -> tuple[int]:
        """The shape of the array, determined by the dimensions."""
        return self.dims.shape()

    def set_values(self, values: np.ndarray):
        """Set the values of the FlodymArray and check if the shape is correct.

        For safety reasons, using scalars or broadcasting smaller arrays is not allowed,
        i.e. the shape of the values must match the shape of the FlodymArray.

        As a less safe but more flexible alternative, you can use e.g. flodym_array.values[...] = 3.5 to set the values directly.
        """
        self.values = values
        self._check_value_format()

    def sum_values(self):
        """Return the sum of all values in the FlodymArray."""
        return np.sum(self.values)

    def sum_values_over(self, sum_over_dims: tuple = ()):
        """Return the sum of the FlodymArray over a given tuple of dimensions.

        Args:
            sum_over_dims (tuple, optional): Tuple of dimension letters to sum over. If not given, no summation is performed and the values array is returned.

        Returns:
            np.ndarray: The partially summed values of the FlodymArray.
        """
        sum_over_dims = self._tuple_to_letters(sum_over_dims)
        result_dims = (o for o in self.dims.letters if o not in sum_over_dims)
        return np.einsum(f"{self.dims.string}->{''.join(result_dims)}", self.values)

    def cast_values_to(self, target_dims: DimensionSet):
        """Cast the values of the FlodymArray to a new set of dimensions.

        Args:
            target_dims (DimensionSet): New dimensions to cast the values to. Must be given as a DimensionSet object, as the new dimensions are otherwise not known to the FlodymArray object.

        Returns:
            np.ndarray: The values of the FlodymArray cast to the new dimensions.
        """
        assert all([d in target_dims.letters for d in self.dims.letters]), (
            "Target of cast must contain all "
            f"dimensions of the object! Source dims '{self.dims.string}' are not all contained in target dims "
            f"'{target_dims.string}'. Maybe use sum_values_to() before casting"
        )
        # safety procedure: order dimensions
        values = np.einsum(
            f"{self.dims.string}->{''.join([d for d in target_dims.letters if d in self.dims.letters])}",
            self.values,
        )
        index = tuple(
            [slice(None) if d in self.dims.letters else np.newaxis for d in target_dims.letters]
        )
        multiple = tuple([1 if d.letter in self.dims.letters else d.len for d in target_dims])
        values = values[index]
        values = np.tile(values, multiple)
        return values

    def cast_to(self, target_dims: DimensionSet):
        """Cast the FlodymArray to a new set of dimensions.

        Args:
            target_dims (DimensionSet): New dimensions to cast the FlodymArray to. Must be given as a DimensionSet object, as the new dimensions are otherwise not known to the FlodymArray object.

        Returns:
            FlodymArray: The FlodymArray cast to the new dimensions.
        """
        return FlodymArray(
            dims=target_dims, values=self.cast_values_to(target_dims), name=self.name
        )

    def sum_values_to(self, result_dims: tuple[str] = ()):
        """Return the values of the FlodymArray partially summed, such that only the dimensions given in the result_dims tuple are left.

        Args:
            result_dims (tuple, optional): Tuple of dimension letters to sum over. If not given, the sum over all dimensions is returned.
        """
        result_dims = self._tuple_to_letters(result_dims)
        return np.einsum(f"{self.dims.string}->{''.join(result_dims)}", self.values)

    def sum_to(self, result_dims: tuple = ()) -> "FlodymArray":
        """Return the FlodymArray summed, such that only the dimensions given in the result_dims tuple are left.

        Args:
            result_dims (tuple, optional): Tuple of the dimensions to sum to. If not given, the sum over all dimensions is returned.

        Returns:
            FlodymArray: FlodymArray object with the summed values and the reduced dimensions.
        """
        result_dims = self._tuple_to_letters(result_dims)
        return FlodymArray(
            dims=self.dims.get_subset(result_dims),
            values=self.sum_values_to(result_dims),
            name=self.name,
        )

    def sum_over(self, sum_over_dims: tuple = ()) -> "FlodymArray":
        """Return the FlodymArray summed over a given tuple of dimensions.

        Args:
            sum_over_dims (tuple, optional): Tuple of dimension letters to sum over. If not given, no summation is performed and the FlodymArray object is returned.

        Returns:
            FlodymArray: FlodymArray object with the summed values and the reduced dimensions.
        """
        sum_over_dims = self._tuple_to_letters(sum_over_dims)
        result_dims = tuple([d for d in self.dims.letters if d not in sum_over_dims])
        return FlodymArray(
            dims=self.dims.get_subset(result_dims),
            values=self.sum_values_over(sum_over_dims),
            name=self.name,
        )

    def _tuple_to_letters(self, dim_tuple: tuple) -> tuple:
        """Ensure that an input dimension tuple is converted to a tuple of dimension letters,
        if e.g. names or Dimension objects are given instead.

        Args:
            dim_tuple (tuple): Tuple of dimensions, which can be given as letters, names or Dimension objects.

        Returns:
            tuple: Tuple of dimension letters.
        """
        return tuple(self._get_dim_letter(item) for item in dim_tuple)

    def _get_dim_letter(self, dim: Union[str, Dimension]) -> str:
        """Get the letter of a dimension, given either the letter or the name of the dimension, or the Dimension object.

        Args:
            dim (Union[str, Dimension]): Dimension letter, name or object.

        Returns:
            str: Dimension letter.
        """
        if isinstance(dim, Dimension):
            return dim.letter
        elif isinstance(dim, str) and dim in self.dims:
            return self.dims[dim].letter
        else:
            raise KeyError(f"Dimension {dim} not found in FlodymArray dims.")

    def _prepare_other(self, other: Union["FlodymArray", int, float]) -> "FlodymArray":
        """If a math operation between a FlodymArray and an int or float is performed, the int or float is converted to a FlodymArray object.
        The following operations are then performed between the two FlodymArray objects.

        Args:
            other (Union[FlodymArray, int, float]): The other object to perform the operation with.

        Returns:
            FlodymArray: The other object converted to a FlodymArray object.
        """
        assert isinstance(other, (FlodymArray, int, float)), (
            "Can only perform operations between two " "FlodymArrays or FlodymArray and scalar."
        )
        if isinstance(other, (int, float)):
            other = FlodymArray(dims=self.dims, values=other * np.ones(self.shape))
        return other

    def __add__(self, other):
        other = self._prepare_other(other)
        dims_out = self.dims.intersect_with(other.dims)
        return FlodymArray(
            dims=dims_out,
            values=self.sum_values_to(dims_out.letters) + other.sum_values_to(dims_out.letters),
        )

    def __sub__(self, other):
        other = self._prepare_other(other)
        dims_out = self.dims.intersect_with(other.dims)
        return FlodymArray(
            dims=dims_out,
            values=self.sum_values_to(dims_out.letters) - other.sum_values_to(dims_out.letters),
        )

    def __mul__(self, other):
        other = self._prepare_other(other)
        dims_out = self.dims.union_with(other.dims)
        values_out = np.einsum(
            f"{self.dims.string},{other.dims.string}->{dims_out.string}", self.values, other.values
        )
        return FlodymArray(dims=dims_out, values=values_out)

    def __truediv__(self, other):
        other = self._prepare_other(other)
        dims_out = self.dims.union_with(other.dims)
        values_out = np.einsum(
            f"{self.dims.string},{other.dims.string}->{dims_out.string}",
            self.values,
            1.0 / other.values,
        )
        return FlodymArray(dims=dims_out, values=values_out)

    def minimum(self, other):
        other = self._prepare_other(other)
        dims_out = self.dims.intersect_with(other.dims)
        values_out = np.minimum(
            self.sum_values_to(dims_out.letters), other.sum_values_to(dims_out.letters)
        )
        return FlodymArray(dims=dims_out, values=values_out)

    def maximum(self, other):
        other = self._prepare_other(other)
        dims_out = self.dims.intersect_with(other.dims)
        values_out = np.maximum(
            self.sum_values_to(dims_out.letters), other.sum_values_to(dims_out.letters)
        )
        return FlodymArray(dims=dims_out, values=values_out)

    def apply(
        self, func: callable, kwargs: dict = {}, inplace: bool = False
    ) -> Optional["FlodymArray"]:
        """Apply a function to the values of the FlodymArray.

        Args:
            func (callable): Function to apply to the values. Must be a function that can be applied to a numpy array, and return a numpy array of the same shape.
            kwargs (dict): Keyword argument dictionary to pass to the function.
            inplace (bool, optional): Whether to apply the function in place. Defaults to False.

        Returns:
            FlodymArray: FlodymArray object with the values transformed by the function.
        """
        if inplace:
            self.values = func(self.values, **kwargs)
            return
        return FlodymArray(dims=self.dims, values=func(self.values, **kwargs))

    def abs(self, inplace: bool = False):
        return self.apply(np.abs, inplace=inplace)

    def sign(self, inplace: bool = False):
        return self.apply(np.sign, inplace=inplace)

    def cumsum(self, dim_letter: str, inplace: bool = False):
        """Calculate the cumulative sum along a dimension.

        Args:
            dim_letter (str): Dimension letter to calculate the cumulative sum along.
            inplace (bool, optional): Whether to apply the cumulative sum in place. Defaults to False.

        Returns:
            FlodymArray: FlodymArray object with the cumulative sum along the given dimension.
        """
        i_axis = self.dims.letters.index(dim_letter)
        return self.apply(np.cumsum, kwargs={"axis": i_axis}, inplace=inplace)

    def __neg__(self):
        return FlodymArray(dims=self.dims, values=-self.values)

    def __abs__(self):
        return FlodymArray(dims=self.dims, values=abs(self.values))

    def __radd__(self, other):
        return self + other

    def __rsub__(self, other):
        return -self + other

    def __rmul__(self, other):
        return self * other

    def __rtruediv__(self, other):
        inv_self = FlodymArray(dims=self.dims, values=1 / self.values)
        return inv_self * other

    def __getitem__(self, keys):
        """Defines what is returned when the object with square brackets stands on the right-hand side of an assignment,
        e.g. foo = foo = bar[{'e': 'C'}] Here, it is solely used for slicing, the the input tot the square brackets must
        be a dictionary defining the slice."""
        return self._sub_array_handler(keys).to_flodym_array()

    def __setitem__(self, keys, item):
        """Defines what is returned when the object with square brackets stands on the left-hand side of an assignment,
        i.e. 'foo[bar] = baz' For allowed values in the square brackets (bar), see the docstring of the SubArrayHandler
        class.

        The RHS (baz) is required here to be a FlodymArray.
        If you want to set the values of a FlodymArray object directly to a numpy array, use the syntax
        'foo.values[...] = bar'."""
        assert isinstance(item, FlodymArray), "Item on RHS of assignment must be a FlodymArray"
        slice_obj = self._sub_array_handler(keys)
        self.values[slice_obj.ids] = item.sum_values_to(slice_obj.dim_letters)
        return

    def to_df(self, index: bool = True, dim_to_columns: str = None) -> pd.DataFrame:
        """Export the FlodymArray to a pandas DataFrame.

        Args:
            index (bool, optional): Whether to include the dimension items as a Multi-Index (True) or as columns of the DataFrame (False). Defaults to True.
            dim_to_columns (str, optional): Name of the dimension the items of which are to form the columns of the DataFrame. If not given, the DataFrame is returned in long format with a single 'value' column.

        Returns:
            pd.DataFrame: DataFrame representation of the FlodymArray.
        """
        multiindex = pd.MultiIndex.from_product([d.items for d in self.dims], names=self.dims.names)
        df = pd.DataFrame({"value": self.values.flatten()})
        df = df.set_index(multiindex)
        if dim_to_columns is not None:
            if dim_to_columns not in self.dims:
                raise ValueError(f"Dimension name {dim_to_columns} not found in flodym_array.dims")
            # transform to name, if given as letter
            dim_to_columns = self.dims[dim_to_columns].name
            df.reset_index(inplace=True)
            index_names = [n for n in self.dims.names if n != dim_to_columns]
            df = df.pivot(index=index_names, columns=dim_to_columns, values="value")
        if not index:
            df.reset_index(inplace=True)
        return df

    def set_values_from_df(self, df_in: pd.DataFrame):
        """Set the values of the FlodymArray from a pandas DataFrame.

        Parameters:
            df (DataFrame): pandas DataFrame containing the values of the FlodymArray.
                Dimensions of the named dim array can be given in DataFrame columns or the index.
                The DataFrame can be in long or wide format, that is there can either be one value column,
                or the value columns are named by items of one FlodymArray dimension.
                If dimension names are not given in the respective index or column, they are inferred from the
                items of the dimensions of the FlodymArray.
                It is advisable to give the dimension names in the DataFrame, as this makes the error messages
                more informative if there are typos in the items or if items are missing.
                Ordering of rows and columns is arbitrary, but the items across each dimension must be given,
                must be complete and exactly match those of the FlodymArray.
                Dimensions with only one item do not need to be given in the DataFrame.
                Supersets of dimensions (i.e. additional values) will lead to an error.
        """
        self.set_values(DataFrameToFlodymDataConverter(df_in, self).target_values)

    def split(self, dim_letter: str) -> dict:
        """Reverse the flodym_array_stack, returns a dictionary of FlodymArray objects
        associated with the item in the dimension that has been split.
        Method can be applied to classes FlodymArray, StockArray, Parameter and Flow.
        """
        return {item: self[{dim_letter: item}] for item in self.dims[dim_letter].items}

    def get_shares_over(self, dim_letters: tuple) -> "FlodymArray":
        """Get shares of the FlodymArray along a tuple of dimensions, indicated by letter."""
        assert all(
            [d in self.dims.letters for d in dim_letters]
        ), "Dimensions to get share of must be in the object"

        if all([d in dim_letters for d in self.dims.letters]):
            return self / self.sum_values()

        return self / self.sum_over(sum_over_dims=dim_letters)


class SubArrayHandler:
    """This class handles subsets of the 'values' numpy array of a FlodymArray object, created by slicing along one or
    several dimensions. It specifies the behavior of `foo[definition] = bar` and `foo = bar[definition]`, where `foo` and `bar`
    are FlodymArray objects. This is done via the `__getitem__` and `__setitem__` methods of the FlodymArray class.

    It returns either

    - a new FlodymArray object (via the `to_flodym_array()` function), or
    - a pointer to a subset of the values array of the parent FlodymArray object, via the `values_pointer` attribute.

    There are several possible syntaxes for the definition of the subset:

    - An ellipsis slice `...` can be used to address all the values of the original FlodymArray object

      *Example:* `foo[...]` addresses all values of the FlodymArray object `foo`.
    - A dictionary can be used to define a subset along one or several dimensions.
      The dictionary has the form `{'dim_letter': 'item_name'}`.

      *Example:* `foo[{'e': 'C'}]` addresses all values where the element is carbon,

      Instead of a single 'item_name', a list of 'item_names' can be passed.

      *Example:* `foo[{'e': 'C', 'r': ['EUR', 'USA']}]` addresses all values where the element is carbon and the region is
      Europe or the USA.
    - Instead of a dictionary, an item name can be passed directly. In this case, the dimension is inferred from the
      item name.
      Throws an error if the item name is not unique, i.e. occurs in more than one dimension.

      *Example:* `foo['C']` addresses all values where the element is carbon

      Several comma-separated item names can be passed, which appear in `__getitem__` and `__setitem__` methods as a tuple.
      Those can either be in the same dimension or in different dimensions.

      *Example:* `foo['C', 'EUR', 'USA']` addresses all values where the element is carbon and the region is Europe or the
      USA.

    Note that does not inherit from FlodymArray, so it is not a FlodymArray object itself.
    However, one can use it to create a FlodymArray object with the `to_flodym_array()` method.
    """

    def __init__(self, flodym_array: FlodymArray, definition):
        self.flodym_array = flodym_array
        self._get_def_dict(definition)
        self.invalid_flodym_array = any(_is_iterable(v) for v in self.def_dict.values())
        self._init_dims_out()
        self._init_ids()

    def _get_def_dict(self, definition):
        if isinstance(definition, type(Ellipsis)):
            self.def_dict = {}
        elif isinstance(definition, dict):
            self.def_dict = definition
        elif isinstance(definition, tuple):
            self.def_dict = self._to_dict_tuple(definition)
        else:
            self.def_dict = {self._get_key_single_item(definition): definition}

    def _get_key_single_item(self, item):
        if isinstance(item, slice):
            raise ValueError(
                "Numpy indexing of FlodymArrays is not supported. Details are given in the FlodymArray class "
                "docstring."
            )
        key = None
        for d in self.flodym_array.dims:
            if item in d.items:
                if key is not None:
                    raise ValueError(
                        f"Ambiguous slicing: Item '{item}' is found in multiple dimensions. Please specify the "
                        "dimension by using a slicing dict instead."
                    )
                key = d.letter
        if key is None:
            raise ValueError(f"Slicing item '{item}' not found in any dimension.")
        return key

    def _to_dict_tuple(self, slice_def) -> dict:
        dict_out = defaultdict(list)
        for item in slice_def:
            key = self._get_key_single_item(item)
            dict_out[key].append(item)
        # if there is only one item along a dimension, convert list to single item
        return {k: v if len(v) > 1 else v[0] for k, v in dict_out.items()}

    @property
    def ids(self):
        """Indices used for slicing the values array."""
        return tuple(self._id_list)

    @property
    def values_pointer(self):
        """Pointer to the subset of the values array of the parent FlodymArray object."""
        return self.flodym_array.values[self.ids]

    def _init_dims_out(self):
        self.dims_out = deepcopy(self.flodym_array.dims)
        for letter, value in self.def_dict.items():
            if isinstance(value, Dimension):
                self.dims_out.replace(letter, value, inplace=True)
            elif not _is_iterable(value):
                self.dims_out.drop(letter, inplace=True)

    @property
    def dim_letters(self):
        """Updated dimension letters, where sliced dimensions with only one item along that direction are removed."""
        return self.dims_out.letters

    def to_flodym_array(self) -> "FlodymArray":
        """Return a FlodymArray object that is a slice of the original FlodymArray object.

        Attention: This creates a new FlodymArray object, which is not linked to the original one.
        """
        if self.invalid_flodym_array:
            raise ValueError(
                "Cannot convert to FlodymArray if there are dimension slices with several items."
                "Use a new dimension object with the subset as values instead"
            )

        return FlodymArray(
            dims=self.dims_out, values=self.values_pointer, name=self.flodym_array.name
        )

    def _init_ids(self):
        """
        - Init the internal list of index slices to slice(None) (i.e. no slicing, keep all items along that dimension)
        - For each dimension that is sliced, get the corresponding item IDs and set the index slice to these IDs.
        """
        self._id_list = [slice(None) for _ in self.flodym_array.dims.letters]
        for dim_letter, item_or_items in self.def_dict.items():
            self._set_ids_single_dim(dim_letter, item_or_items)

    def _set_ids_single_dim(self, dim_letter, item_or_items):
        """Given either a single item name or a list of item names, return the corresponding item IDs, along one
        dimension 'dim_letter'."""
        if isinstance(item_or_items, Dimension):
            if item_or_items.is_subset(self.flodym_array.dims[dim_letter]):
                items_ids = [
                    self._get_single_item_id(dim_letter, item) for item in item_or_items.items
                ]
            else:
                raise ValueError(
                    "Dimension item given in array index must be a subset of the dimension it replaces"
                )
        elif _is_iterable(item_or_items):
            items_ids = [self._get_single_item_id(dim_letter, item) for item in item_or_items]
        else:
            items_ids = self._get_single_item_id(dim_letter, item_or_items)  # single item
        self._id_list[self.flodym_array.dims.index(dim_letter)] = items_ids

    def _get_single_item_id(self, dim_letter, item_name):
        return self.flodym_array.dims[dim_letter].items.index(item_name)


class Flow(FlodymArray):
    """The values of Flow objects are the main computed outcome of the MFA system.
    A Flow object connects two :py:class:`Process` objects.
    The name of the Flow object is set as a combination of the names of the two processes it connects.

    Flow is a subclass of :py:class:`FlodymArray`, so most of its methods are inherited.

    **Example**

        >>> from flodym import DimensionSet, Flow, Process
        >>> goods = Dimension(name='Good', letter='g', items=['Car', 'Bus', 'Bicycle'])
        >>> time = Dimension(name='Time', letter='t', items=[1990, 2000, 2010, 2020, 2030])
        >>> dimensions = DimensionSet([goods, time])
        >>> fabrication = Process(name='fabrication', id=2)
        >>> use = Process(name='use', id=3)
        >>> flow = Flow(from_process='fabrication', to_process='use', dims=dimensions)

    In the above example, we did not pass any values when initialising the Flow instance,
    and these would get filled with zeros.
    See the validation (filling) method in :py:class:`FlodymArray`.
    """

    model_config = ConfigDict(protected_namespaces=())

    from_process: Process
    """Process from which the flow originates."""
    to_process: Process
    """Process to which the flow goes."""

    @property
    def from_process_id(self):
        """ID of the process from which the flow originates."""
        return self.from_process.id

    @property
    def to_process_id(self):
        """ID of the process to which the flow goes."""
        return self.to_process.id


class StockArray(FlodymArray):
    """Stocks allow accumulation of material at a process, i.e. between two flows.

    StockArray inherits all its functionality from :py:class:`FlodymArray`.
    StockArray's are used in the :py:class:`flodym.Stock` for the inflow, outflow and stock.
    """

    pass


class Parameter(FlodymArray):
    """Parameter's can be used when defining the :py:meth:`flodym.MFASystem.compute` of a specific MFA system,
    to quantify the links between specific :py:class:`flodym.Stock` and :py:class:`Flow` objects,
    for example as the share of flows that go into one branch when the flow splits at a process.

    Parameter inherits all its functionality from :py:class:`FlodymArray`.
    """

    pass



"""
ODYM example by Stefan Pauliuk, adapted for flodym.
Same as example2.
"""

import numpy as np

from flodym import (
    Dimension,
    DimensionSet,
    FlodymArray,
    Parameter,
    FlowDefinition,
    StockDefinition,
    MFASystem,
    SimpleFlowDrivenStock,
    make_empty_flows,
    make_empty_stocks,
    make_processes,
)


class ExampleMFA(MFASystem):
    def compute(self):
        self.flows["sysenv => shredder"][...] = (
            self.parameters["eol machines"] * self.parameters["composition eol machines"]
        )
        self.flows["sysenv => demolition"][...] = (
            self.parameters["eol buildings"] * self.parameters["composition eol buildings"]
        )
        self.flows["shredder => remelting"][...] = (
            self.flows["sysenv => shredder"] * self.parameters["shredder yield"]
        )
        self.flows["shredder => sysenv"][...] = self.flows["sysenv => shredder"] * (
            1 - self.parameters["shredder yield"]
        )
        self.flows["demolition => remelting"][...] = (
            self.flows["sysenv => demolition"] * self.parameters["demolition yield"]
        )
        self.flows["demolition => landfills"][...] = self.flows["sysenv => demolition"] * (
            1 - self.parameters["demolition yield"]
        )
        self.flows["remelting => sysenv"][...] = (
            self.flows["shredder => remelting"] + self.flows["demolition => remelting"]
        ) * self.parameters["remelting yield"]
        self.flows["remelting => slag piles"][...] = (
            self.flows["shredder => remelting"] + self.flows["demolition => remelting"]
        ) * (1 - self.parameters["remelting yield"])
        self.stocks["landfills"].inflow[...] = self.flows["demolition => landfills"]
        self.stocks["landfills"].compute()
        self.stocks["slag piles"].inflow[...] = self.flows["shredder => remelting"]
        self.stocks["slag piles"].compute()


def _get_dims() -> DimensionSet:
    return DimensionSet(
        dim_list=[
            Dimension(letter="t", name="Time", dtype=int, items=list(range(1980, 2011))),
            Dimension(letter="e", name="Material", dtype=str, items=["Fe", "Cu", "Mn"]),
        ]
    )


def get_example_array() -> FlodymArray:
    dims = _get_dims()
    return FlodymArray(dims=dims, name="Example Array", values=np.random.random((31, 3)))


def get_example_mfa() -> ExampleMFA:
    dims = _get_dims()

    parameters = {
        "eol machines": Parameter(
            name="eol machines", dims=dims[("t",)], values=np.arange(15.5, 31.0, 0.5)
        ),
        "eol buildings": Parameter(
            name="eol buildings", dims=dims[("t",)], values=np.arange(120, 368, 8)
        ),
        "composition eol machines": Parameter(
            name="composition eol machines", dims=dims[("e",)], values=np.array([0.8, 0.15, 0.05])
        ),
        "composition eol buildings": Parameter(
            name="composition eol buildings",
            dims=dims[("e",)],
            values=np.array([0.95, 0.045, 0.005]),
        ),
        "shredder yield": Parameter(
            name="shredder yield", dims=dims[("e",)], values=np.array([0.92, 0.1, 0.92])
        ),
        "demolition yield": Parameter(
            name="demolition yield", dims=dims[("e",)], values=np.array([0.95, 0.02, 0.95])
        ),
        "remelting yield": Parameter(
            name="remelting yield", dims=dims[("e",)], values=np.array([0.96, 1.0, 0.5])
        ),
    }

    process_names = [
        "sysenv",
        "shredder",
        "demolition",
        "remelting",
        "landfills",
        "slag piles",
    ]
    processes = make_processes(process_names)

    flow_definitions = [
        FlowDefinition(
            from_process_name="sysenv", to_process_name="shredder", dim_letters=("t", "e")
        ),
        FlowDefinition(
            from_process_name="sysenv", to_process_name="demolition", dim_letters=("t", "e")
        ),
        FlowDefinition(
            from_process_name="shredder",
            to_process_name="remelting",
            dim_letters=("t", "e"),
        ),  # scrap type 1
        FlowDefinition(
            from_process_name="shredder", to_process_name="sysenv", dim_letters=("t", "e")
        ),  # shredder residue
        FlowDefinition(
            from_process_name="demolition",
            to_process_name="remelting",
            dim_letters=("t", "e"),
        ),  # scrap type 2
        FlowDefinition(
            from_process_name="demolition",
            to_process_name="landfills",
            dim_letters=("t", "e"),
        ),  # loss
        FlowDefinition(
            from_process_name="remelting",
            to_process_name="slag piles",
            dim_letters=("t", "e"),
        ),  # secondary steel
        FlowDefinition(
            from_process_name="remelting", to_process_name="sysenv", dim_letters=("t", "e")
        ),  # slag
    ]
    flows = make_empty_flows(processes, flow_definitions, dims)

    stock_definitions = [
        StockDefinition(
            name="landfills",
            process="landfills",
            dim_letters=("t", "e"),
            subclass=SimpleFlowDrivenStock,
        ),
        StockDefinition(
            name="slag piles",
            process="slag piles",
            dim_letters=("t", "e"),
            subclass=SimpleFlowDrivenStock,
        ),
    ]
    stocks = make_empty_stocks(stock_definitions, processes, dims)

    return ExampleMFA(
        dims=dims, parameters=parameters, processes=processes, flows=flows, stocks=stocks
    )



from matplotlib import pyplot as plt
from plotly import graph_objects as go, colors as plc
from plotly.subplots import make_subplots
import numpy as np
from pydantic import BaseModel as PydanticBaseModel, model_validator, ConfigDict
from typing import Any, Optional, Union
from abc import ABC, abstractmethod

from ..flodym_arrays import FlodymArray
from ..dimensions import DimensionSet
from .helper import CustomNameDisplayer


class ArrayPlotter(CustomNameDisplayer, ABC, PydanticBaseModel):
    """
    Abstract base class for array plotting classes. Subclasses exist for pyplot and plotly.
    Mostly recommended for plotting multi-dimensional arrays, where subplots and multiple lines are needed.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow", protected_namespaces=())

    array: FlodymArray
    """Values to plot, usually a Flow or Stock; sliced or summed along excess dimensions."""
    intra_line_dim: str
    """Name of the dimension along which lines are plotted (if no x_array is given, this is also the x-axis)."""
    x_array: Optional[Union[FlodymArray, None]] = None
    """Array with x-values for each line. Must have the same dimensions as array, or a subset of them. If None, the intra_line_dim is used as x-axis."""
    subplot_dim: str = None
    """Name of the dimension by which to split the array into subplots. If None, the array is plotted in a single subplot."""
    linecolor_dim: str = None
    """Name of the dimension along which to split the array into several lines within each subplot. If None, only one line is plotted per subplot."""
    fig: Any = None
    """Pre-existing figure to plot on. If None, a new figure is created."""
    line_label: str = None
    """Custom label for the line. If None, the respective item along linecolor_dim is used as label."""
    xlabel: str = None
    """Custom label for the x-axis. If None, the name of the x_array or intra_line_dim is used."""
    ylabel: str = None
    """Custom label for the y-axis. If None, the name of the array is used."""
    title: str = None
    """Title of the plot, if desired."""
    color_map: list[str] = None
    """List of colors to use for the lines. If None, a default color map is used."""
    chart_type: str = "line"
    """Type of chart to plot. Can be 'line', 'scatter', or 'area'."""
    line_type: str = "solid"
    """Type of line to plot. Can be 'solid', 'dashed', 'dotted', or 'dashdot'."""
    suppress_legend: bool = False
    """If True, the legend is not shown."""
    __pydantic_extra__: dict[str, Any]

    @model_validator(mode="after")
    def check_chart_type(self):
        if self.chart_type not in self._allowed_chart_types:
            raise ValueError("chart_type must be either 'line' or 'scatter'.")
        return self

    @model_validator(mode="after")
    def check_line_type(self):
        if self.line_type not in self._allowed_line_types:
            allowed_changes = {"dash": "dashed", "dot": "dotted", "dashed": "dash", "dotted": "dot"}
            if (
                self.line_type in allowed_changes
                and allowed_changes[self.line_type] in self._allowed_line_types
            ):
                self.line_type = allowed_changes[self.line_type]
            else:
                raise ValueError(f"line_type must be one of {self._allowed_line_types}.")
        if self.chart_type != "line" and self.line_type != "solid":
            raise ValueError("line_type is only applicable to chart_type 'line'.")
        return self

    @property
    def _allowed_line_types(self):
        raise NotImplementedError

    @property
    def _allowed_chart_types(self):
        return ["line", "area", "scatter"]

    @model_validator(mode="after")
    def check_colors(self):
        if self.linecolor_dim is not None and self.line_label is not None:
            raise ValueError(
                "If linecolor_dim is given, several lines are plotted. In this case, lines are labeled"
                "by items along that dimension, and a line_label must not be given."
            )
        return self

    @model_validator(mode="after")
    def check_dims(self):
        dim_attributes = [self.linecolor_dim, self.subplot_dim, self.intra_line_dim]
        dim_attribute_names = ["linecolor_dim", "subplot_dim", "intra_line_dim"]
        for attr_name, dim in zip(dim_attribute_names, dim_attributes):
            if dim is not None and dim not in self.array.dims.names:
                raise ValueError(f"Dimension {dim} given in {attr_name} not in array dimensions.")
        specified_dim_attributes = [d for d in dim_attributes if d is not None]
        excess_dims = set(self.array.dims.names) - set(specified_dim_attributes)
        if excess_dims:
            raise ValueError(
                "All dimensions of passed array must be given exactly once. Either as subplot_dim, linecolor_dim, or intra_line_dim."
                + f"Excess dimensions: {', '.join(excess_dims)}; "
                + "Sum or slice array along these dims before passing it to the plotter."
            )
        if self.x_array is not None:
            if any(d not in self.array.dims.names for d in self.x_array.dims.names):
                raise ValueError(
                    "x_array must have the same dimensions as array, or a subset of them."
                )
        return self

    def plot(self, save_path: str = None, do_show: bool = False):
        self._fill_fig()
        subplots_array, subplots_x_array = self._prepare_arrays()
        self._plot_all_subplots(subplots_array, subplots_x_array)
        self.plot_legend()
        if self.title is not None:
            self.set_title()
        if save_path is not None:
            self.save(save_path)
        if do_show:
            self.show()
        return self.fig

    def _prepare_arrays(self) -> tuple[dict[str, FlodymArray], dict[str, FlodymArray]]:
        self._get_x_array_like_value_array()
        subplots_array = self._dict_of_slices(self.array, self.subplot_dim).values()
        subplots_x_array = self._dict_of_slices(self.x_array, self.subplot_dim).values()
        return subplots_array, subplots_x_array

    @staticmethod
    def _dict_of_slices(array: FlodymArray, dim_name_to_slice) -> dict[str, FlodymArray]:
        if dim_name_to_slice is not None:
            slice_letter = array.dims[dim_name_to_slice].letter
            return array.split(slice_letter)
        else:
            return {None: array}

    def _plot_all_subplots(self, subplotlist_array, subplotlist_x_array):
        for i_subplot, (array_subplot, x_array_subplot) in enumerate(
            zip(subplotlist_array, subplotlist_x_array)
        ):
            self._plot_subplot(i_subplot=i_subplot, array=array_subplot, x_array=x_array_subplot)
            self._label_subplot(i_subplot=i_subplot)

    def _fill_fig(self):
        if self.fig is not None:  # already filled from input argument
            self.nx, self.ny = self._get_nx_ny()
            self.subplot_titles = None
            return
        if self.subplot_dim is None:
            self.nx, self.ny = 1, 1
            self.subplot_titles = None
        else:
            n_subplots = self.array.dims[self.subplot_dim].len
            self.nx, self.ny = self._get_nx_ny_from_n(n_subplots)
            self.subplot_titles = [
                f"{self.display_name(self.subplot_dim)}={self.display_name(item)}"
                for item in self.array.dims[self.subplot_dim].items
            ]
        self.fig = self.get_fig()

    def _get_x_array_like_value_array(self):
        if self.x_array is None:
            x_dim_obj = self.array.dims[self.intra_line_dim]
            x_dimset = DimensionSet(dim_list=[x_dim_obj])
            self.x_array = FlodymArray(
                dims=x_dimset, values=np.array(x_dim_obj.items), name=self.intra_line_dim
            )
        self.x_array = self.x_array.cast_to(self.array.dims)

    def _plot_subplot(self, i_subplot: int, array: FlodymArray, x_array: FlodymArray):
        linedict_array = self._dict_of_slices(array, self.linecolor_dim)
        linedict_x_array = self._dict_of_slices(x_array, self.linecolor_dim)
        prev_y = None
        for i_line, (array_line, x_array_line, name_line) in enumerate(
            zip(linedict_array.values(), linedict_x_array.values(), linedict_array.keys())
        ):
            label = self.line_label if self.line_label is not None else self.display_name(name_line)
            assert array_line.dims.names == (self.intra_line_dim,), (
                "All dimensions of array must be given exactly once. Either as x_dim / subplot_dim / linecolor_dim, or in "
                "slice_dict or summed_dims."
            )
            self.add_line(i_subplot, x_array_line.values, array_line.values, prev_y, label, i_line)
            prev_y = array_line.values

    def _label_subplot(self, i_subplot: int):
        if self.subplot_titles is not None:
            self.set_subplot_title(i_subplot, self.subplot_titles[i_subplot])
        xlabel = self.xlabel if self.xlabel is not None else self.display_name(self.x_array.name)
        if xlabel != "unnamed":
            self.set_xlabel(i_subplot, xlabel)
        ylabel = self.ylabel if self.ylabel is not None else self.display_name(self.array.name)
        if ylabel != "unnamed":
            self.set_ylabel(i_subplot, ylabel)

    def _index2d(self, i_subplot):
        return i_subplot // self.nx, i_subplot % self.nx

    @staticmethod
    def _get_nx_ny_from_n(n):
        nx = int(np.ceil(np.sqrt(n)))
        ny = int(np.ceil(n / nx))
        return nx, ny

    @abstractmethod
    def save(self, save_path: str = None):
        raise NotImplementedError

    @abstractmethod
    def show(self):
        raise NotImplementedError

    @abstractmethod
    def _get_nx_ny(self):
        raise NotImplementedError

    @abstractmethod
    def get_fig(self):
        raise NotImplementedError

    @abstractmethod
    def set_xlabel(self, index, label):
        raise NotImplementedError

    @abstractmethod
    def set_ylabel(self, index, label):
        raise NotImplementedError

    @abstractmethod
    def set_subplot_title(self, index, title):
        raise NotImplementedError

    @abstractmethod
    def add_line(self, index, x, y, prev_y, label, i_line):
        raise NotImplementedError

    @abstractmethod
    def plot_legend(self):
        raise NotImplementedError

    @abstractmethod
    def set_title(self):
        raise NotImplementedError


class PyplotArrayPlotter(ArrayPlotter):

    fig: plt.Figure = None
    """A previously created pyplot figure object, for adding lines to an existing figure.
    If None, a new figure is created.
    """

    def save(self, save_path: str = None, **kwargs):
        self.fig.savefig(save_path, **kwargs)

    def show(self):
        self.fig.show()

    def get_fig(self):
        fig, _ = plt.subplots(self.nx, self.ny)
        return fig

    def _get_nx_ny(self):
        return self._get_nx_ny_from_n(len(self.ax))

    def set_xlabel(self, i_subplot, label):
        self.ax[i_subplot].set_xlabel(label)

    def set_ylabel(self, i_subplot, label):
        self.ax[i_subplot].set_ylabel(label)

    def set_subplot_title(self, i_subplot, title):
        self.ax[i_subplot].set_title(title)

    @property
    def _allowed_line_types(self):
        return ["solid", "dashed", "dotted", "dashdot"]

    def add_line(self, i_subplot, x, y, prev_y, label, i_line):
        common_dict = {}
        if not self.suppress_legend:
            common_dict["label"] = label
        if self.color_map is not None:
            common_dict["color"] = self.color_map[i_line]

        if self.chart_type == "line":
            self.ax[i_subplot].plot(x, y, linestyle=self.line_type, **common_dict)
        elif self.chart_type == "scatter":
            self.ax[i_subplot].scatter(x, y, **common_dict)
        elif self.chart_type == "area":
            args = [x, y] if prev_y is None else [x, y, prev_y]
            self.ax[i_subplot].fill_between(*args, **common_dict)

    def plot_legend(self):
        handles, labels = self.ax[0].get_legend_handles_labels()
        self.fig.legend(handles, labels, loc="lower center")

    def set_title(self):
        self.fig.suptitle(self.title)

    @property
    def ax(self) -> plt.Axes:
        return self.fig.axes


class PlotlyArrayPlotter(ArrayPlotter):

    fig: go.Figure = None
    """A previously created plotly figure object, for adding lines to an existing figure.
    If None, a new figure is created.
    """
    color_map: list[str] = plc.qualitative.Dark24
    """List of colors to use for the lines. If None, a default color map is used."""

    def save(self, save_path: str = None, **kwargs):
        self.fig.write_image(save_path, **kwargs)

    def show(self):
        self.fig.show()

    def get_fig(self):
        fig = make_subplots(self.ny, self.nx, subplot_titles=self.subplot_titles)
        return fig

    def _fill_fig(self):
        super()._fill_fig()
        self.n_previous_lines = getattr(self.fig, "_n_lines_flodym", 0)
        n_current_lines = (
            len(self.array.dims[self.linecolor_dim].items) if self.linecolor_dim is not None else 1
        )
        self.fig._n_lines_flodym = self.n_previous_lines + n_current_lines

    def _get_nx_ny(self):
        grid_ref = self.fig._validate_get_grid_ref()
        nrows = len(grid_ref)
        ncols = len(grid_ref[0])
        return ncols, nrows

    def row(self, i_subplot):
        return self._index2d(i_subplot)[0] + 1

    def col(self, i_subplot):
        return self._index2d(i_subplot)[1] + 1

    def set_xlabel(self, i_subplot, label):
        self.fig.update_xaxes(title_text=label, row=self.row(i_subplot), col=self.col(i_subplot))

    def set_ylabel(self, i_subplot, label):
        self.fig.update_yaxes(title_text=label, row=self.row(i_subplot), col=self.col(i_subplot))

    def set_subplot_title(self, index, title):
        pass  # already set in make_subplots

    @property
    def _allowed_line_types(self):
        return ["solid", "dash", "dot", "dashdot"]

    def add_line(self, i_subplot, x, y, prev_y, label, i_line):
        i_color = i_line + self.n_previous_lines
        color = self.color_map[i_color]
        common_dict = dict(
            x=x,
            y=y,
            name=label,
            showlegend=i_subplot == 0 and not self.suppress_legend,
        )
        if self.chart_type == "line":
            trace = go.Scatter(
                **common_dict,
                line=dict(color=color, dash=self.line_type),
            )
        elif self.chart_type == "scatter":
            trace = go.Scatter(
                **common_dict,
                mode="markers",
                marker=dict(color=color),
            )
        elif self.chart_type == "area":
            trace = go.Scatter(
                **common_dict,
                fill="tozeroy" if prev_y is None else "tonexty",
                fillcolor=color,
                line=dict(color=color),
            )
        else:
            raise ValueError("chart_type must be either 'line' or 'scatter'.")
        self.fig.add_trace(
            trace=trace,
            row=self.row(i_subplot),
            col=self.col(i_subplot),
        )

    def plot_legend(self):
        pass

    def set_title(self):
        self.fig.update_layout(title={"text": self.title})



"""Home to a base class for MFA systems.

Specific MFA models can be built that inherit from this class.
"""

import logging
from typing import Dict, Optional

import numpy as np
from pydantic import BaseModel as PydanticBaseModel, ConfigDict

from .mfa_definition import MFADefinition
from .dimensions import DimensionSet
from .flodym_arrays import Flow, Parameter, FlodymArray
from .stocks import Stock
from .processes import Process, make_processes
from .stock_helper import make_empty_stocks
from .flow_helper import make_empty_flows
from .data_reader import (
    DataReader,
    CompoundDataReader,
    CSVDimensionReader,
    CSVParameterReader,
    ExcelDimensionReader,
    ExcelParameterReader,
)


class MFASystem(PydanticBaseModel):
    """An MFASystem class handles the calculation of a Material Flow Analysis system, which
    consists of a set of processes, flows, stocks defined over a set of dimensions.
    For the concrete definition of the system, a subclass of MFASystem must be implemented.

    **Example**
    Define your MFA System:

        >>> from flodym import MFASystem
        >>> class CustomMFA(MFASystem):
        >>>     def compute(self):
        >>>         # do some computations on the CustomMFA attributes: stocks and flows

    MFA flows, stocks and parameters are defined as instances of subclasses of :py:class:`flodym.FlodymArray`.
    Dimensions are managed with the :py:class:`flodym.Dimension` and :py:class:`flodym.DimensionSet`.
    """

    model_config = ConfigDict(protected_namespaces=(), extra="allow")

    dims: DimensionSet
    """All dimensions that appear in the MFA system."""
    parameters: Dict[str, Parameter]
    """The parameters of the MFA system,
    as a dictionary mapping the names of the MFA system parameters to the parameters themselves.
    """
    processes: Dict[str, Process]
    """The processes of the MFA system, i.e. the nodes of the MFA system graph,
    as a dictionary mapping the names of the MFA system processes to the processes themselves.
    """
    flows: Dict[str, Flow]
    """The flows of the MFA system, i.e. the edges of the MFA system graph,
    as a dictionary mapping the names of the MFA system flows to the flows themselves.
    """
    stocks: Optional[Dict[str, Stock]] = {}
    """The stocks of the MFA system,
    as a dictionary mapping the names of the MFA system stocks to the stocks themselves.
    """

    @classmethod
    def from_data_reader(cls, definition: MFADefinition, data_reader: DataReader) -> "MFASystem":
        """Define and set up the MFA system and load all required data.
        Initialises stocks and flows with all zero values."""
        dims = data_reader.read_dimensions(definition.dimensions)
        parameters = data_reader.read_parameters(definition.parameters, dims=dims)
        processes = make_processes(definition.processes)
        flows = make_empty_flows(processes=processes, flow_definitions=definition.flows, dims=dims)
        stocks = make_empty_stocks(
            processes=processes, stock_definitions=definition.stocks, dims=dims
        )
        return cls(
            dims=dims,
            parameters=parameters,
            processes=processes,
            flows=flows,
            stocks=stocks,
        )

    @classmethod
    def from_csv(
        cls,
        definition: MFADefinition,
        dimension_files: dict,
        parameter_files: dict,
    ):
        """Define and set up the MFA system and load all required data from CSV files.
        Initialises stocks and flows with all zero values.

        See :py:class:`flodym.CSVDimensionReader` and
        :py:class:`flodym.CSVParameterReader`, and :py:meth:`flodym.FlodymArray.from_df` for expected
        format.

        :param definition: The MFA definition object
        :param dimension_files: A dictionary mapping dimension names to CSV files
        :param parameter_files: A dictionary mapping parameter names to CSV files
        """

        dimension_reader = CSVDimensionReader(
            dimension_files=dimension_files,
        )
        parameter_reader = CSVParameterReader(
            parameter_files=parameter_files,
        )
        data_reader = CompoundDataReader(
            dimension_reader=dimension_reader,
            parameter_reader=parameter_reader,
        )
        return cls.from_data_reader(definition, data_reader)

    @classmethod
    def from_excel(
        cls,
        definition: MFADefinition,
        dimension_files: dict,
        parameter_files: dict,
        dimension_sheets: dict = None,
        parameter_sheets: dict = None,
    ):
        """Define and set up the MFA system and load all required data from Excel files.
        Initialises stocks and flows with all zero values.
        Builds a CompoundDataReader from Excel readers, and calls the from_data_reader class method.

        See :py:class:`flodym.ExcelDimensionReader`,
        :py:class:`flodym.ExcelParameterReader`, and
        :py:meth:`flodym.FlodymArray.from_df` for expected format.

        :param definition: The MFA definition object
        :param dimension_files: A dictionary mapping dimension names to Excel files
        :param parameter_files: A dictionary mapping parameter names to Excel files
        :param dimension_sheets: A dictionary mapping dimension names to sheet names in the Excel files
        :param parameter_sheets: A dictionary mapping parameter names to sheet names in the Excel files
        """
        dimension_reader = ExcelDimensionReader(
            dimension_files=dimension_files,
            dimension_sheets=dimension_sheets,
        )
        parameter_reader = ExcelParameterReader(
            parameter_files=parameter_files,
            parameter_sheets=parameter_sheets,
        )
        data_reader = CompoundDataReader(
            dimension_reader=dimension_reader,
            parameter_reader=parameter_reader,
        )
        return cls.from_data_reader(definition, data_reader)

    def compute(self):
        """Perform all computations for the MFA system.
        This method must be implemented in a subclass of MFASystem.
        """
        raise NotImplementedError(
            "The compute method must be implemented in a subclass of MFASystem if it is to be used."
        )

    def get_new_array(self, dim_letters: tuple = None, **kwargs) -> FlodymArray:
        """get a new FlodymArray object.

        :param dim_letters: tuple of dimension letters to include in the new FlodymArray. If None, all dimensions are included.
        :param kwargs: keyword arguments to pass to the FlodymArray constructor.
        """
        dims = self.dims.get_subset(dim_letters)
        return FlodymArray(dims=dims, **kwargs)

    def _get_mass_contributions(self):
        """List all contributions to the mass balance of each process:
        - all flows entering are positive
        - all flows leaving are negative
        - the stock change of the process
        """
        contributions = {p: [] for p in self.processes.keys()}

        # Add flows to mass balance
        for flow in self.flows.values():
            contributions[flow.from_process.name].append(-flow)  # Subtract flow from start process
            contributions[flow.to_process.name].append(flow)  # Add flow to end process

        # Add stock changes to the mass balance
        for stock in self.stocks.values():
            if stock.process is None:  # not connected to a process
                continue
            # add/subtract stock changes to processes
            contributions[stock.process.name].append(-stock.inflow)
            contributions[stock.process.name].append(stock.outflow)
            # add/subtract stock changes in system boundary for mass balance of whole system
            contributions["sysenv"].append(stock.inflow)
            contributions["sysenv"].append(-stock.outflow)

        return contributions

    def _get_mass_balance(self, contributions: dict = {}):
        """Calculate the mass balance for each process, by summing the contributions.
        The sum returns a :py:class:`flodym.FlodymArray`,
        with the dimensions common to all contributions.
        """
        if not contributions:
            contributions = self._get_mass_contributions()
        return {p_name: sum(parts) for p_name, parts in contributions.items()}

    def _get_mass_totals(self, contributions: dict = {}):
        """Calculate the total mass of a process by summing the absolute values of all
        the contributions.
        """
        if not contributions:
            contributions = self._get_mass_contributions()
        return {
            p_name: sum([abs(part) for part in parts]) for p_name, parts in contributions.items()
        }

    def _get_relative_mass_balance(self, epsilon=1e-9):
        """Determines a relative mass balance for each process of the MFA system,
        by dividing the mass balances by the mass totals.
        """
        mass_contributions = self._get_mass_contributions()
        balances = self._get_mass_balance(contributions=mass_contributions)
        totals = self._get_mass_totals(contributions=mass_contributions)

        relative_balance = {
            p_name: (balances[p_name] / (totals[p_name] + epsilon)).values
            for p_name in self.processes
        }
        return relative_balance

    def check_mass_balance(self, tolerance=1e-4):
        """Compute mass balance, and check whether it is within a certain tolerance.
        Throw an error if it isn't."""

        # returns array with dim [t, process, e]
        relative_balance = self._get_relative_mass_balance()  # assume no error if total sum is 0
        id_failed = {p_name: np.any(rb > tolerance) for p_name, rb in relative_balance.items()}
        messages_failed = [
            f"{p_name} ({np.max(relative_balance[p_name])*100:.2f}% error)"
            for p_name in self.processes.keys()
            if id_failed[p_name]
        ]
        if any(id_failed.values()):
            raise RuntimeError(
                f"Error, Mass Balance fails for processes {', '.join(messages_failed)}"
            )
        else:
            logging.info(
                f"Success - Mass balance of {self.__class__.__name__} object is consistent!"
            )
        return



"""Home to definition classes.

These are used when defining the MFA system, and can be used to check the input data
and put it into ojects with the desired properties.
"""

import numpy as np
from pydantic import (
    BaseModel as PydanticBaseModel,
    AliasChoices,
    Field,
    field_validator,
    model_validator,
    ConfigDict,
)
from typing import List, Optional


class DimensionDefinition(PydanticBaseModel):
    """Define the model dimensions.

    **Examples**

        >>> from flodym import DimensionDefinition
        >>> time_definition = DimensionDefinition(name='Time', letter='t', dtype=int)
        >>> region_definition = DimensionDefinition(name='Region', letter='r', dtype=str)
    """

    model_config = ConfigDict(protected_namespaces=())

    name: str = Field(..., min_length=2)
    letter: str = Field(
        ..., min_length=1, max_length=1, validation_alias=AliasChoices("letter", "dim_letter")
    )
    dtype: type


class DefinitionWithDimLetters(PydanticBaseModel):
    """Base class for any definition that contains dimesnsion letters."""

    model_config = ConfigDict(protected_namespaces=())

    dim_letters: tuple
    """letters of the dimensions that the object is defined on"""

    @field_validator("dim_letters", mode="before")
    def check_dimensions(cls, v):
        for letter in v:
            if (not isinstance(letter, str)) or (len(letter) != 1):
                raise ValueError("Dimensions must be defined using single digit letters")
        return v


class FlowDefinition(DefinitionWithDimLetters):
    """Define the model flows.

    **Examples**

        >>> from flodym import FlowDefinition
        >>> flow_one = FlowDefinition(from_process_name='fabrication', to_process_name='use', dim_letters=('r', 't'))
        >>> flow_two = FlowDefinition(from_process_name='use', to_process_name='end_of_life', dim_letters=('r', 't'))

    These are then used in the :py:class:MFADefinition, for creating a custom MFA System.
    """

    from_process_name: str = Field(
        validation_alias=AliasChoices("from_process_name", "from_process")
    )
    """Process from which the flow originates."""
    to_process_name: str = Field(validation_alias=AliasChoices("to_process_name", "to_process"))
    """Process to which the flow goes."""
    name_override: Optional[str] = None
    """Optional name for the flow. Will be generated from the connecting process names if not provided."""


class StockDefinition(DefinitionWithDimLetters):
    """Define the model stocks."""

    name: str = "undefined stock"
    """Name of the stock."""
    process_name: Optional[str] = Field(
        default=None, validation_alias=AliasChoices("process", "process_name")
    )
    """Name of the process to which the stock is connected."""
    time_letter: str = "t"
    """Letter of the time dimension, to ensure it's the first appearing in dim_letters."""
    subclass: type
    """type of stock. Can be any found in :py:data:`flodym.stocks`."""
    lifetime_model_class: Optional[type] = None
    """Lifetime model used for the stock. Only needed if type is not simple_flow_driven.
    Available lifetime models can be found in :py:data:`flodym.lifetime_models`.
    """

    @model_validator(mode="after")
    def check_lifetime_model(self):
        if (
            self.lifetime_model_class is not None
            and "lifetime_model" not in self.subclass.__fields__
        ):
            raise ValueError(f"Lifetime model is given, but not used in subclass {self.subclass}.")
        elif self.lifetime_model_class is None and "lifetime_model" in self.subclass.__fields__:
            raise ValueError(
                f"Lifetime model class must be part of definition for given subclass {self.subclass}"
            )
        return self


class ParameterDefinition(DefinitionWithDimLetters):
    """Define the model parameters."""

    name: str
    """Name of the parameter."""


class MFADefinition(PydanticBaseModel):
    """All the information needed to define an MFA system, compiled of lists of definition objects."""

    model_config = ConfigDict(protected_namespaces=())

    dimensions: List[DimensionDefinition]
    """List of definitions of dimensions used in the model."""
    processes: List[str]
    """List of process names used in the model."""
    flows: List[FlowDefinition]
    """List of definitions of flows used in the model."""
    stocks: List[StockDefinition]
    """List of definitions of stocks used in the model."""
    parameters: List[ParameterDefinition]
    """List of definitions of parameters used in the model."""

    @model_validator(mode="after")
    def check_dimension_letters(self):
        """Check that dimension letters used for flows, stocks and parameters are part of the defined dimensions."""
        defined_dim_letters = [dd.letter for dd in self.dimensions]
        for item in self.flows + self.stocks + self.parameters:
            correct_dims = [letter in defined_dim_letters for letter in item.dim_letters]
            if not np.all(correct_dims):
                raise ValueError(f"Undefined dimension in {item}")
        return self



from typing import List

from pydantic import BaseModel as PydanticBaseModel, model_validator


class Process(PydanticBaseModel):
    """Processes serve as nodes for the MFA system layout definition.
    Flows are defined between two processes. Stocks are connected to a process.
    Processes do not contain values themselves.

    Processes get an ID by the order they are defined in the :py:attribute::`MFASystem.definition`.
    The process with ID 0 necessarily contains everything outside the system boundary.
    It has to be named 'sysenv'.
    """

    name: str
    """Name of the process."""
    id: int
    """ID of the process."""

    @model_validator(mode="after")
    def check_id0(self):
        """Ensure that the process with ID 0 is named 'sysenv'."""
        if self.id == 0 and self.name != "sysenv":
            raise ValueError(
                "The process with ID 0 must be named 'sysenv', as it contains everything outside the system boundary."
            )
        return self


def make_processes(definitions: List[str]) -> dict[str, Process]:
    """Create a dictionary of processes from a list of process names."""
    return {name: Process(name=name, id=id) for id, name in enumerate(definitions)}
